from __future__ import annotations

import numpy as np

from dxtbx.model.experiment_list import Experiment, ExperimentList
from dxtbx.serialize import load
from scitbx import matrix

from dials.algorithms import shoebox
from dials.algorithms.profile_model.gaussian_rs import Model
from dials.array_family import flex


def predict_reflections(sequence, crystal):
    # Get models from the sequence
    beam = sequence.get_beam()
    detector = sequence.get_detector()
    gonio = sequence.get_goniometer()
    scan = sequence.get_scan()

    sigma_b = beam.get_sigma_divergence(deg=True)
    sigma_m = crystal.get_mosaicity(deg=True)

    exlist = ExperimentList()
    exlist.append(
        Experiment(
            imageset=sequence,
            beam=beam,
            detector=detector,
            goniometer=gonio,
            scan=scan,
            crystal=crystal,
            profile=Model(None, 3, sigma_b, sigma_m, deg=True),
        )
    )

    predicted = flex.reflection_table.from_predictions(exlist[0])
    predicted["id"] = flex.int(len(predicted), 0)
    predicted.compute_bbox(exlist)

    # Find overlapping reflections
    overlaps = shoebox.find_overlapping(predicted["bbox"])

    # Return the reflections and overlaps
    return predicted, overlaps


def test(dials_data):
    # Load the sequence and crystal
    sequence = load.imageset(
        dials_data("centroid_test_data", pathlib=True) / "sweep.json"
    )
    crystal = load.crystal(
        str(dials_data("centroid_test_data", pathlib=True) / "crystal.json")
    )

    # Get models from the sequence
    detector = sequence.get_detector()

    # Get the reflections and overlaps
    reflections, adjacency_list = predict_reflections(sequence, crystal)
    reflections["shoebox"] = flex.shoebox(reflections["panel"], reflections["bbox"])
    reflections["shoebox"].allocate_with_value(shoebox.MaskCode.Valid)

    # If the adjacency list is given, then create the reflection mask
    assert len(detector) == 1
    image_size = detector[0].get_image_size()
    shoeboxes = reflections["shoebox"]
    coords = reflections["xyzcal.px"]
    shoebox_masker = shoebox.MaskOverlapping()
    shoebox_masker(shoeboxes, coords, adjacency_list)

    # Loop through all edges
    overlapping = []
    for e in adjacency_list.edges():
        v1, v2 = adjacency_list.source(e), adjacency_list.target(e)
        overlapping.append(v1)
        overlapping.append(v2)

    # Ensure elements are unique
    overlapping = set(overlapping)

    # Ensure we have some overlaps
    assert len(overlapping) > 0

    # Get all non-overlapping reflections
    all_r = set(range(len(reflections)))
    non_overlapping = all_r.difference(overlapping)

    # Run the tests
    tst_non_overlapping(reflections, non_overlapping, detector[0].get_image_size())
    tst_overlapping(reflections, overlapping, adjacency_list, image_size)


def tst_non_overlapping(reflections, non_overlapping, image_size):
    """Ensure non-overlapping reflections have all their values 1."""
    # Check that all elements in non_overlapping masks are 1
    shoeboxes = reflections["shoebox"]
    for i in non_overlapping:
        mask = shoeboxes[i].mask
        assert mask.all_eq(shoebox.MaskCode.Valid)


def tst_overlapping(reflections, overlapping, adjacency_list, image_size):
    """Ensure masks for overlapping reflections are set properly."""
    # Loop through all overlaps
    shoeboxes = reflections["shoebox"]
    coord = reflections["xyzcal.px"]
    for i in overlapping:
        r1 = shoeboxes[i]
        bbox_1 = r1.bbox
        r1_coord = matrix.col(coord[i])

        # Create a mask that we expect
        r1_size = (bbox_1[5] - bbox_1[4], bbox_1[3] - bbox_1[2], bbox_1[1] - bbox_1[0])
        expected_mask = np.zeros(shape=r1_size, dtype=np.int32)
        expected_mask[:, :, :] = shoebox.MaskCode.Valid

        # Loop through all reflections which this reflection overlaps
        for j in adjacency_list.adjacent_vertices(i):
            r2 = shoeboxes[j]
            bbox_2 = r2.bbox
            r2_coord = matrix.col(coord[j])

            # Get bounding box of intersection
            bbox_3 = (
                max(bbox_1[0], bbox_2[0]),
                min(bbox_1[1], bbox_2[1]),
                max(bbox_1[2], bbox_2[2]),
                min(bbox_1[3], bbox_2[3]),
                max(bbox_1[4], bbox_2[4]),
                min(bbox_1[5], bbox_2[5]),
            )

            # Check intersection is valid
            assert bbox_3[0] < bbox_3[1]
            assert bbox_3[2] < bbox_3[3]
            assert bbox_3[4] < bbox_3[5]

            # Get the coordinates are all mask values
            mask_coord = []
            for k in range(bbox_3[4], bbox_3[5]):
                for j in range(bbox_3[2], bbox_3[3]):
                    for i in range(bbox_3[0], bbox_3[1]):
                        mask_coord.append(matrix.col((i + 0.5, j + 0.5, k + 0.5)))

            def dist(a, m):
                return np.array([(a - b).length() for b in m])

            # Find the indices in the intersection area where r2 is closer to
            # the point than r1
            ind = np.where(dist(r1_coord, mask_coord) > dist(r2_coord, mask_coord))[0]

            # Set the mask values for r1 where r2 is closer to 0
            k0, k1 = bbox_3[4] - bbox_1[4], bbox_3[5] - bbox_1[4]
            j0, j1 = bbox_3[2] - bbox_1[2], bbox_3[3] - bbox_1[2]
            i0, i1 = bbox_3[0] - bbox_1[0], bbox_3[1] - bbox_1[0]
            intersect_mask = expected_mask[k0:k1, j0:j1, i0:i1]
            intersect_mask_1d = intersect_mask.reshape(-1)
            intersect_mask_1d[ind] = 0
            intersect_mask[:, :] = intersect_mask_1d.reshape(intersect_mask.shape)
            expected_mask[k0:k1, j0:j1, i0:i1] = intersect_mask

        # Check the masks are the same
        calculated_mask = r1.mask.as_numpy_array()
        assert np.all(calculated_mask == expected_mask)



"""This module defines some useful header functions for CBF files."""

from __future__ import annotations

import numpy as np
import pycbf

from scitbx import matrix


def compute_central_rotation_matrix(gonio):
    """Compute the central rotation matrix

    :param gonio: The goniometer struct
    :returns: The central rotation matrix
    """
    x = gonio.rotate_vector(0.5, 1, 0, 0)
    y = gonio.rotate_vector(0.5, 0, 1, 0)
    z = gonio.rotate_vector(0.5, 0, 0, 1)
    R = matrix.sqr(x + y + z).transpose()
    return R


def get_image(cbf_handle, category="array_data", column="data", row=0, element=0):
    """Read an image from a CBF file

    This function is a bit of a hack - I'm not sure what the general structure
    of a CBF file is like but for the data I have, it works. Reads an image
    from the location specified in the CBF file, otherwise raises an exception.

    :param cbf_handle: The handle to the CBF file
    :param category: Category in which the image is contained
    :param column: Column in which the image is contained
    :param row: Row in which image is contained
    :param element: Element in which image is contained
    :returns: An array of image data
    """
    # Find the given category, column and row
    cbf_handle.find_category(category)
    cbf_handle.find_column(column)
    cbf_handle.select_row(row)

    # Check the type of the element to ensure it's a binary
    # otherwise raise an exception
    if "bnry" in cbf_handle.get_typeofvalue():

        # Read the image data into an array
        image_string = cbf_handle.get_integerarray_as_string()
        image = np.fromstring(image_string, np.int32)

        # Get the size of the image data (either 2d or 3d)
        image_size = cbf_handle.get_image_size(element)

        # Resize the image
        image.shape = image_size

    else:
        raise TypeError(f"{category}:{column}:{row}:{element} is not an image")

    # Return the image
    return image


def get_image_volume(cbf_paths):
    """Load the image volume from the list of cbf_paths. The list of paths is
    assumed to be is order from 1->n.

    :param cbf_paths: The list of cbf files
    :param width The width (xsize) of the volume
    :param height The height (ysize) of the volume
    :returns: The 3D volume array
    """
    # Read the first image and get the size
    cbf_handle = pycbf.cbf_handle_struct()
    cbf_handle.read_file(cbf_paths[0], pycbf.MSG_DIGEST)
    image = get_image(cbf_handle)
    height, width = image.shape

    # Initialise the image volume
    num_slices = len(cbf_paths)
    volume = np.zeros(shape=(num_slices, height, width), dtype=np.int32)
    volume[0, :, :] = image

    # For each CBF file, read the image and put into the image volume
    for i, filename in enumerate(cbf_paths[1:]):
        cbf_handle = pycbf.cbf_handle_struct()
        cbf_handle.read_file(filename, pycbf.MSG_DIGEST)
        volume[i + 1, :, :] = get_image(cbf_handle)

    # Return the image volume
    return volume


def search_for_image_volume(search_path):
    """Load the CBF image volume

    Args:
        search_path: The CBF file search path

    Returns:
        The image volume
    """
    from glob import glob

    # Load the CBF image volume
    cbf_path = sorted(glob(search_path))
    return get_image_volume(cbf_path)



from __future__ import annotations

import math

import pytest

from dxtbx.model.experiment_list import ExperimentListFactory

from dials.algorithms.refinement.reflection_manager import ReflectionManager
from dials.array_family import flex


def test_scan_margin(dials_data):

    # Use 4 scan data for this test
    data_dir = dials_data("l_cysteine_dials_output", pathlib=True)
    experiments = ExperimentListFactory.from_json_file(
        data_dir / "indexed.expt", check_format=False
    )
    reflections = flex.reflection_table.from_file(data_dir / "indexed.refl")
    orig_phi = reflections["xyzobs.mm.value"].parts()[2]

    # Reflection Manager works on predictions, but this dataset has none, so
    # need to set the predictions flag
    reflections.set_flags(
        flex.bool(len(reflections), True), reflections.flags.predicted
    )

    # Create a reflection manager without trimming scan margins
    refman = ReflectionManager(reflections, experiments)
    refman.finalise()
    refs1 = refman.get_matches()
    phi1 = refs1["xyzobs.mm.value"].parts()[2]

    # Create a reflection manager with 1 degree width scan margins
    margin = 1.0
    refman2 = ReflectionManager(reflections, experiments, scan_margin=margin)
    refman2.finalise()
    refs2 = refman2.get_matches()
    phi2 = refs2["xyzobs.mm.value"].parts()[2]

    # Check zero scan margins do not trim
    assert min(orig_phi) == min(phi1)
    assert max(orig_phi) == max(phi1)

    # Check 1 degree scan margin trims approximately 1 degree
    assert min(phi2) == pytest.approx(min(phi1) + math.radians(margin), abs=1e-3)
    assert max(phi2) == pytest.approx(max(phi1) - math.radians(margin), abs=1e-3)



from __future__ import annotations

import numpy as np

from dials.algorithms.symmetry.cosym import plots


def test_plot_coords():
    coords = np.array([0.5, 0.5, 0, 1, 0, 1, 1, 0, 1, 0]).reshape(5, 2)
    labels = np.array([-1, 0, 0, 1, 1])
    d = plots.plot_coords(coords, labels=labels)
    assert "cosym_coordinates" in d
    assert set(d["cosym_coordinates"]) == {"data", "help", "layout"}
    assert (
        d["cosym_coordinates"]["data"][0]["marker"]["color"]
        == "rgb(0.000000,0.000000,0.000000)"
    )
    assert d["cosym_coordinates"]["data"][0]["x"] == [0.5]
    assert d["cosym_coordinates"]["data"][0]["y"] == [0.5]
    assert d["cosym_coordinates"]["data"][1]["x"] == [0.0, 0.0]
    assert d["cosym_coordinates"]["data"][1]["y"] == [1.0, 1.0]
    assert d["cosym_coordinates"]["data"][2]["x"] == [1.0, 1.0]
    assert d["cosym_coordinates"]["data"][2]["y"] == [0.0, 0.0]


def test_plot_rij_histogram():
    rij_matrix = np.random.rand(16)
    d = plots.plot_rij_histogram(rij_matrix)
    assert "cosym_rij_histogram" in d
    assert sum(d["cosym_rij_histogram"]["data"][0]["y"]) == 16



from __future__ import annotations

import math

from dials.array_family import flex


def spot_counts_per_image_plot(reflections, **kwargs):
    if reflections.size() == 0:
        return "\n"

    x, y, z = reflections["xyzobs.px.value"].parts()
    return flex_histogram(z, **kwargs)


def flex_histogram(z, char="*", width=60, height=10):
    assert isinstance(char, str)
    assert len(char) == 1

    # with open('list.json', 'w') as fh:
    #   json.dump(sorted(list(z)), fh)

    min_z = flex.min(z)
    max_z = flex.max(z)

    epsilon = flex.double(n % 2 for n in range(len(z)))
    epsilon = (epsilon / 2) - 0.25
    z += epsilon

    # image numbers to display on x-axis label
    xmin = math.ceil(min_z)
    xmax = math.ceil(max_z)

    # estimate the total number of images
    image_count = xmax - xmin + 1
    if image_count <= 1:
        return f"{len(z)} spots found on 1 image"

    # determine histogram width
    width = min(image_count, width)
    z_step = (max_z - min_z) / width

    # bin all spots
    counts = flex.double()
    seen_spots = 0
    for i in range(1, width):
        z_bound = min_z + (i * z_step)
        spots = (z < z_bound).count(True)
        counts.append(spots - seen_spots)
        seen_spots = spots
    spots = (z >= z_bound).count(True)
    counts.append(spots)

    max_count = flex.max(counts)
    total_counts = flex.sum(counts)
    assert total_counts == len(
        z
    ), "Only found %d out of %d reflections for histogram" % (total_counts, len(z))
    counts *= height / max_count
    counts = counts.iround()

    rows = [
        "%i spots found on %i images (max %i / bin)"
        % (total_counts, image_count, max_count)
    ]

    for i in range(height, 0, -1):
        row = []
        for j, c in enumerate(counts):
            if c > (i - 1):
                row.append(char)
            else:
                row.append(" ")
        rows.append("".join(row))

    padding = width - len(str(xmin)) - len(str(xmax))
    rows.append(
        "%i%s%i"
        % (
            xmin,
            (" " if padding < 7 else "image").center(padding) if padding > 0 else "",
            xmax,
        )
    )
    return "\n".join(rows)



"""Managed reflection prediction for refinement.

* ScansRayPredictor adapts DIALS prediction for use in refinement, by keeping
  up to date with the current model geometry
* StillsRayPredictor predicts reflections without a goniometer, under
  the naive assumption that the relp is already in reflecting position
"""


from __future__ import annotations

from math import pi

from scitbx.array_family import flex

from dials.algorithms.spot_prediction import ScanStaticRayPredictor
from dials.algorithms.spot_prediction import ScanStaticReflectionPredictor as sc
from dials.algorithms.spot_prediction import ScanVaryingReflectionPredictor as sv
from dials.algorithms.spot_prediction import StillsReflectionPredictor as st

# constants
TWO_PI = 2.0 * pi


class ScansRayPredictor:
    """
    Predict for a relp based on the current states of models of the
    experimental geometry. This is a wrapper for DIALS' C++
    RayPredictor class, which does the real work. This class keeps track
    of the experimental geometry, and instantiates a RayPredictor when
    required.
    """

    def __init__(self, experiments, sequence_range=(0, 2.0 * pi)):
        """Construct by linking to instances of experimental model classes"""

        self._experiments = experiments
        self._sequence_range = sequence_range

    def __call__(self, hkl, experiment_id=0, UB=None):
        """
        Solve the prediction formula for the reflecting angle phi.

        If UB is given, override the contained crystal model. This is
        for use in refinement with time-varying crystal parameters
        """

        e = self._experiments[experiment_id]
        ray_predictor = ScanStaticRayPredictor(
            e.beam.get_s0(),
            e.goniometer.get_rotation_axis_datum(),
            e.goniometer.get_fixed_rotation(),
            e.goniometer.get_setting_rotation(),
            self._sequence_range,
        )

        UB_ = UB if UB else e.crystal.get_A()

        rays = ray_predictor(hkl, UB_)

        return rays


class ExperimentsPredictor:
    """
    Predict for relps based on the current states of models of the experimental
    geometry. This version manages multiple experiments, selecting the correct
    predictor in each case.
    """

    def __init__(self, experiments):
        """Construct by linking to instances of experimental model classes"""

        self._experiments = experiments

    def __call__(self, reflections):
        """Predict for all reflections at the current model geometry"""

        for iexp, e in enumerate(self._experiments):

            # select the reflections for this experiment only
            sel = reflections["id"] == iexp
            refs = reflections.select(sel)

            self._predict_one_experiment(e, refs)

            # write predictions back to overall reflections
            reflections.set_selected(sel, refs)

        reflections = self._post_prediction(reflections)

        return reflections

    def _predict_one_experiment(self, experiment, reflections):

        raise NotImplementedError()

    def _post_prediction(self, reflections):
        """Perform tasks on the whole reflection list after prediction before
        returning."""

        return reflections


class ScansExperimentsPredictor(ExperimentsPredictor):
    def _predict_one_experiment(self, experiment, reflections):

        # scan-varying
        if "ub_matrix" in reflections:
            predictor = sv(experiment)
            UB = reflections["ub_matrix"]
            s0 = reflections["s0_vector"]
            dmat = reflections["d_matrix"]
            Smat = reflections["S_matrix"]
            predictor.for_reflection_table(reflections, UB, s0, dmat, Smat)
        # scan static
        else:
            predictor = sc(experiment)
            UB = experiment.crystal.get_A()
            predictor.for_reflection_table(reflections, UB)

    def _post_prediction(self, reflections):

        if "xyzobs.mm.value" in reflections:
            reflections = self._match_full_turns(reflections)

        return reflections

    def _match_full_turns(self, reflections):
        """Modify the calculated phi values so that they match the full rotation
        from zero taken from the the observations, rather than being modulo 2*pi."""

        x_obs, y_obs, phi_obs = reflections["xyzobs.mm.value"].parts()
        x_calc, y_calc, phi_calc = reflections["xyzcal.mm"].parts()
        resid = phi_calc - (flex.fmod_positive(phi_obs, TWO_PI))
        # ensure this is the smaller of two possibilities
        resid = flex.fmod_positive((resid + pi), TWO_PI) - pi
        phi_calc = phi_obs + resid
        reflections["xyzcal.mm"] = flex.vec3_double(x_calc, y_calc, phi_calc)

        # Update xyzcal.px with the correct z_px values in keeping with above
        for iexp, e in enumerate(self._experiments):
            sel = reflections["id"] == iexp
            x_px, y_px, z_px = reflections["xyzcal.px"].select(sel).parts()
            scan = e.scan
            if scan is not None:
                z_px = scan.get_array_index_from_angle(phi_calc.select(sel), deg=False)
            else:
                # must be a still image, z centroid not meaningful
                z_px = phi_calc.select(sel)
            xyzcal_px = flex.vec3_double(x_px, y_px, z_px)
            reflections["xyzcal.px"].set_selected(sel, xyzcal_px)

        return reflections


class StillsExperimentsPredictor(ExperimentsPredictor):

    spherical_relp_model = False

    def _predict_one_experiment(self, experiment, reflections):

        predictor = st(experiment, spherical_relp=self.spherical_relp_model)
        UB = experiment.crystal.get_A()
        predictor.for_reflection_table(reflections, UB)


class ExperimentsPredictorFactory:
    @staticmethod
    def from_experiments(experiments, force_stills=False, spherical_relp=False):

        # Determine whether or not to use a stills predictor
        if not force_stills:
            for exp in experiments:
                if exp.goniometer is None:
                    force_stills = True
                    break

        # Construct the predictor
        if force_stills:
            predictor = StillsExperimentsPredictor(experiments)
            predictor.spherical_relp_model = spherical_relp
        else:
            predictor = ScansExperimentsPredictor(experiments)

        return predictor



from __future__ import annotations

import os

import sphinx.directives.code as code

"""
Add a directive to literal-include from generated DIALS output logs
and an associated configuration parameter pointing to the location of said logs
"""


def setup(app):
    app.add_config_value("dials_logs", None, "html")
    app.add_directive("dials_tutorial_include", DialsTutorialInclude)
    return {"parallel_read_safe": True}


class DialsTutorialInclude(code.LiteralInclude):
    """Behaves like LiteralInclude but looks for files relative to the known
    DIALS output log directory."""

    def run(self) -> code.List[code.nodes.Node]:
        document = self.state.document
        if not document.settings.file_insertion_enabled:
            return [
                document.reporter.warning("File insertion disabled", line=self.lineno)
            ]
        # convert options['diff'] to absolute path
        if "diff" in self.options:
            _, path = self.env.relfn2path(self.options["diff"])
            self.options["diff"] = path

        try:
            location = self.state_machine.get_source_and_line(self.lineno)
            filename = os.path.join(self.config.dials_logs, self.arguments[0])
            self.env.note_dependency(filename)

            reader = code.LiteralIncludeReader(filename, self.options, self.config)
            text, lines = reader.read(location=location)

            retnode = code.nodes.literal_block(text, text, source=filename)
            self.set_source_info(retnode)
            if self.options.get("diff"):  # if diff is set, set udiff
                retnode["language"] = "udiff"
            elif "language" in self.options:
                retnode["language"] = self.options["language"]
            retnode["linenos"] = (
                "linenos" in self.options
                or "lineno-start" in self.options
                or "lineno-match" in self.options
            )
            retnode["classes"] += self.options.get("class", [])
            extra_args = retnode["highlight_args"] = {}
            if "emphasize-lines" in self.options:
                hl_lines = code.parselinenos(self.options["emphasize-lines"], lines)
                if any(i >= lines for i in hl_lines):
                    code.logger.warning(
                        code.__("line number spec is out of range(1-%d): %r")
                        % (lines, self.options["emphasize-lines"]),
                        location=location,
                    )
                extra_args["hl_lines"] = [x + 1 for x in hl_lines if x < lines]
            extra_args["linenostart"] = reader.lineno_start

            if "caption" in self.options:
                caption = self.options["caption"] or self.arguments[0]
                retnode = code.container_wrapper(self, retnode, caption)

            # retnode will be note_implicit_target that is linked from caption and numref.
            # when options['name'] is provided, it should be primary ID.
            self.add_name(retnode)

            return [retnode]
        except Exception as exc:
            print(exc)
            return [document.reporter.warning(str(exc), line=self.lineno)]



from __future__ import annotations

import socket
import subprocess
import sys
import time
import timeit
import urllib.request
from xml.dom import minidom

import procrunner
import pytest


@pytest.fixture
def server_port(tmp_path) -> int:
    """Fixture to load a find_spots_server server"""
    if sys.hexversion >= 0x3080000 and sys.platform == "darwin":
        pytest.skip("find_spots server known to be broken on MacOS with Python 3.8+")
    if sys.platform == "win32":
        pytest.skip("find_spots server is not supported on Windows")

    # Find a free port to run the server on
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.bind(("", 0))
        host, port = sock.getsockname()
    # Start the server
    server_command = ["dials.find_spots_server", f"port={port}", "nproc=3"]
    p = subprocess.Popen(server_command, cwd=tmp_path)
    wait_for_server(port)
    yield port
    p.terminate()
    p.wait(timeout=3)
    p.kill()


def test_server_return_codes(dials_data, server_port):
    try:
        first_file = sorted(
            dials_data("centroid_test_data", pathlib=True).glob("*.cbf")
        )[0]
        response = urllib.request.urlopen(
            f"http://127.0.0.1:{server_port}/{first_file}"
        )
        assert response.code == 200
        with pytest.raises(urllib.error.HTTPError):
            urllib.request.urlopen(f"http://127.0.0.1:{server_port}/some/junk/filename")
    finally:
        result = procrunner.run(
            ["dials.find_spots_client", f"port={server_port}", "stop"]
        )
        assert not result.returncode and not result.stderr


def test_find_spots_server_client(dials_data, tmp_path, server_port):
    filenames = sorted(dials_data("centroid_test_data", pathlib=True).glob("*.cbf"))
    try:
        exercise_client(server_port=server_port, filenames=filenames)
    finally:
        result = procrunner.run(
            ["dials.find_spots_client", f"port={server_port}", "stop"]
        )
        assert not result.returncode and not result.stderr


def wait_for_server(port, max_wait=20):
    print(f"Waiting up to {max_wait} seconds for server to start")
    server_ok = False
    start_time = timeit.default_timer()
    max_time = start_time + max_wait
    while (timeit.default_timer() < max_time) and not server_ok:
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.connect(("127.0.0.1", port))
            server_ok = True
        except OSError as e:
            if (e.errno != 111) and (e.errno != 61):
                raise
            # ignore connection failures (111 connection refused on linux; 61 connection refused on mac)
            time.sleep(0.1)
    if not server_ok:
        raise Exception("Server failed to start after %d seconds" % max_wait)
    print(
        "dials.find_spots_server up after %f seconds"
        % (timeit.default_timer() - start_time)
    )


def exercise_client(server_port, filenames):
    assert filenames
    client_command = [
        "dials.find_spots_client",
        f"port={server_port}",
        "min_spot_size=3",
        "algorithm=dispersion",
        "nproc=1",
        filenames[0],
    ]

    index_client_command = client_command + [
        "index=True",
        "indexing.method=fft1d",
        "max_refine=10",
    ]
    print(index_client_command)
    result = procrunner.run(index_client_command)
    assert not result.returncode and not result.stderr
    out = f"<document>{result.stdout}</document>"

    xmldoc = minidom.parseString(out)
    assert len(xmldoc.getElementsByTagName("image")) == 1
    assert len(xmldoc.getElementsByTagName("spot_count")) == 1
    assert len(xmldoc.getElementsByTagName("spot_count_no_ice")) == 1
    assert len(xmldoc.getElementsByTagName("d_min")) == 1
    assert len(xmldoc.getElementsByTagName("total_intensity")) == 1
    assert len(xmldoc.getElementsByTagName("unit_cell")) == 1
    assert len(xmldoc.getElementsByTagName("n_indexed")) == 1
    assert len(xmldoc.getElementsByTagName("fraction_indexed")) == 1

    unit_cell = [
        float(f)
        for f in xmldoc.getElementsByTagName("unit_cell")[0].childNodes[0].data.split()
    ]

    assert unit_cell == pytest.approx(
        [39.90, 42.67, 42.37, 89.89, 90.10, 90.13], abs=1e-1
    )

    client_command = client_command + filenames[1:]
    result = procrunner.run(client_command)
    assert not result.returncode and not result.stderr
    out = f"<document>{result.stdout}</document>"

    xmldoc = minidom.parseString(out)
    images = xmldoc.getElementsByTagName("image")
    assert len(images) == 9
    spot_counts = sorted(
        int(node.childNodes[0].data)
        for node in xmldoc.getElementsByTagName("spot_count")
    )
    assert spot_counts == sorted([203, 196, 205, 209, 195, 205, 203, 207, 189])
    spot_counts_no_ice = sorted(
        int(node.childNodes[0].data)
        for node in xmldoc.getElementsByTagName("spot_count_no_ice")
    )
    assert spot_counts_no_ice == sorted([169, 171, 175, 176, 177, 184, 193, 195, 196])
    d_min = sorted(
        float(node.childNodes[0].data) for node in xmldoc.getElementsByTagName("d_min")
    )
    assert d_min == sorted([1.45, 1.47, 1.55, 1.55, 1.56, 1.59, 1.61, 1.61, 1.64])



from __future__ import annotations

import os
import re
from datetime import datetime
from typing import Any

import requests


def github_api_get(url: str, **kwargs) -> Any:
    """
    Fetch a GitHub API url, with authentication if available.

    If an environment variable GITHUB_TOKEN is set, thi will be used
    when accessing the releases API.
    """
    args = dict(kwargs)
    headers = {
        "Accept": "application/vnd.github.v3+json",
        **args.pop("headers", {}),
    }
    # If we've set an environment variable, use the token
    if token := os.environ.get("GITHUB_TOKEN", None):
        headers["Authorization"] = f"token {token}"
    req = requests.get(url, headers=headers, **args)
    req.raise_for_status()
    return req.json()


def _download_button(text, version, link):
    print(f"  {version} {text} -> {link}")
    return f".. button::\n   :text: DIALS {version} {text}\n   :link: {link}\n\n"


def _sort_tag_version(release_data: dict) -> tuple:
    """Sort a list of releases by tag version."""
    return tuple(int(x) for x in release_data["tag_name"].lstrip("v").split("."))


def _release_asset_buttons(release_info: dict) -> list[str]:
    """Generate the reST markup for the download buttons."""
    assets = {}
    for a in release_info.get("assets", []):
        tag = re.search("dials-v([^-]+)-([^-]+)-([^-]+)-(.+)", a["name"])
        if tag:
            shortname = tag.group(4)
            version = ".".join(tag.group(1, 2, 3))
            last_update = datetime.strptime(
                a["updated_at"], "%Y-%m-%dT%H:%M:%SZ"
            )  # - datetime(1970,1,1)).total_seconds()
            if shortname not in assets or assets[shortname][0] < last_update:
                assets[shortname] = (
                    last_update,
                    version,
                    a.get("browser_download_url"),
                )

    long_names = {
        "macosx.pkg": "Mac installer",
        "macosx.tar.gz": "Mac tar archive",
        "linux-x86_64.tar.xz": "Linux installer",
    }

    return [
        _download_button(long_names.get(asset, asset), _version, link)
        for asset, (_, _version, link) in assets.items()
    ]


if __name__ == "__main__":
    release_file = os.path.join(
        os.path.dirname(__file__), "installation.stable_release"
    )
    print("Checking DIALS release status")
    releases_info = sorted(
        github_api_get("https://api.github.com/repos/dials/dials/releases"),
        reverse=True,
        key=_sort_tag_version,
    )

    latest_release = releases_info[0]
    print(f"Latest release is: {latest_release['name']} ({latest_release['tag_name']})")

    with open(release_file, "w") as release:
        caption = f"Stable Release: {latest_release['name']}"
        release.write(
            f"""
{caption}
{'=' * len(caption)}

The current stable release can be downloaded from `Github <{latest_release['html_url']}>`_,
where you can also find further release notes.

""".lstrip()
        )
        release.write("".join(sorted(_release_asset_buttons(latest_release))))



"""
Classes that take in a scaler and minimisation parameters and
return the scale factors and derivatives of the scale factors w.r.t.
the parameters
"""

from __future__ import annotations

from scitbx import sparse

from dials.array_family import flex
from dials_scaling_ext import row_multiply


class RefinerCalculator:
    """Class that takes in a scaling_apm and calculates the scale factors
    and derivatives for minimisation."""

    @staticmethod
    def _calc_component_scales_derivatives(apm, block_id):
        """Calculate the scales and derivatives for all components for a given
        block, returning each as a list of values from the components."""
        scales = []
        derivatives = []
        for component in apm.components.values():
            sdc = component["object"].calculate_scales_and_derivatives(block_id)
            scales.append(sdc[0])
            derivatives.append(sdc[1])
        return scales, derivatives

    @staticmethod
    def _calculate_scale_factors(apm, block_id, scales):
        """Calculate the overall scale factor for each reflection from individual
        components."""
        if not scales:
            return apm.constant_g_values[block_id]  # needs to return to set in Ih_table
        multiplied_scale_factors = flex.double(scales[0].size(), 1.0)
        for s in scales:
            multiplied_scale_factors *= s
        if apm.constant_g_values:
            multiplied_scale_factors *= apm.constant_g_values[block_id]
        return multiplied_scale_factors

    @staticmethod
    def _calculate_derivatives(apm, block_id, scales, derivatives_list):
        """Calculate the derivatives matrix."""
        if not scales:
            return sparse.matrix(0, 0)
        if len(scales) == 1:
            # only one active parameter, so don't need to chain rule any derivatives
            # for block_id in range(len(apm.n_obs)):
            return derivatives_list[0]
        derivatives = sparse.matrix(apm.n_obs[block_id], apm.n_active_params)
        col_idx = 0
        for i, d in enumerate(derivatives_list):
            scale_multipliers = flex.double(apm.n_obs[block_id], 1.0)
            for j, s1 in enumerate(scales):
                if i != j:
                    scale_multipliers *= s1
            if apm.constant_g_values:
                scale_multipliers *= apm.constant_g_values[block_id]
            next_deriv = row_multiply(d, scale_multipliers)
            derivatives.assign_block(next_deriv, 0, col_idx)
            col_idx += d.n_cols
        return derivatives

    @classmethod
    def calculate_scales_and_derivatives(cls, apm, block_id):
        """Calculate scale factors and derivatives for minimisation."""
        scales, derivatives = cls._calc_component_scales_derivatives(apm, block_id)
        return (
            cls._calculate_scale_factors(apm, block_id, scales),
            cls._calculate_derivatives(apm, block_id, scales, derivatives),
        )



from __future__ import annotations

import os
from pathlib import Path

import procrunner
import pytest

import dxtbx
from dxtbx.format.FormatCBFCspad import FormatCBFCspadInMemory
from dxtbx.imageset import ImageSet, ImageSetData, MemReader
from dxtbx.model.experiment_list import ExperimentListFactory
from libtbx.phil import parse

from dials.array_family import flex
from dials.command_line.stills_process import Processor, phil_scope

cspad_cbf_in_memory_phil = """
dispatch.squash_errors = False
spotfinder {
  filter.min_spot_size=2
  threshold.dispersion.gain=25
  threshold.dispersion.global_threshold=100
}
indexing {
  known_symmetry {
    space_group = P6122
    unit_cell = 92.9 92.9 130.4 90 90 120
  }
  refinement_protocol.d_min_start=1.7
  stills.refine_candidates_with_known_symmetry=True
}
"""

sacla_phil = """
dispatch.squash_errors = True
dispatch.coset = True
input.reference_geometry=%s
indexing {
  known_symmetry {
    space_group = P43212
    unit_cell = 78.9 78.9 38.1 90 90 90
  }
  refinement_protocol.d_min_start = 2.2
  stills.refine_candidates_with_known_symmetry=True
}
spotfinder {
  filter.min_spot_size = 2
}
refinement {
  parameterisation {
    detector.fix_list = Dist,Tau1
  }
}
profile {
  gaussian_rs {
    centroid_definition = com
  }
}
output.composite_output = True
"""


@pytest.mark.xfel
@pytest.mark.parametrize("composite_output", [True, False])
def test_cspad_cbf_in_memory(dials_regression, tmp_path, composite_output):
    # Check the data files for this test exist
    image_path = Path(
        dials_regression,
        "image_examples",
        "LCLS_cspad_nexus",
        "idx-20130301060858801.cbf",
    )
    assert image_path.is_file()

    tmp_path.joinpath("process_lcls.phil").write_text(cspad_cbf_in_memory_phil)

    params = phil_scope.fetch(parse(file_name=tmp_path / "process_lcls.phil")).extract()
    params.output.experiments_filename = None
    params.output.composite_output = composite_output
    cwd = Path.cwd()
    try:
        os.chdir(tmp_path)
        if composite_output:
            processor = Processor(params, composite_tag="memtest")
        else:
            processor = Processor(params)
        mem_img = dxtbx.load(image_path)
        raw_data = mem_img.get_raw_data()  # cache the raw data to prevent swig errors
        mem_img = FormatCBFCspadInMemory(mem_img._cbf_handle)
        mem_img._raw_data = raw_data
        mem_img._cbf_handle = None  # drop the file handle to prevent swig errors
        imgset = ImageSet(ImageSetData(MemReader([mem_img]), None))
        imgset.set_beam(mem_img.get_beam())
        imgset.set_detector(mem_img.get_detector())
        experiments = ExperimentListFactory.from_imageset_and_crystal(imgset, None)
        processor.process_experiments(
            "20130301060858801", experiments
        )  # index/integrate the image
        if composite_output:
            processor.finalize()
            result = "idx-memtest_integrated.refl"
        else:
            result = "idx-20130301060858801_integrated.refl"
    finally:
        os.chdir(cwd)

    n_refls = list(
        range(140, 152)
    )  # large ranges to handle platform-specific differences
    table = flex.reflection_table.from_file(tmp_path / result)
    assert len(table) in n_refls, len(table)
    assert "id" in table
    assert (table["id"] == 0).count(False) == 0


@pytest.mark.xfel
@pytest.mark.parametrize("control_flags", [("use_mpi"), (), ("known_orientations")])
def test_sacla_h5(dials_data, tmp_path, control_flags, in_memory=False):
    use_mpi = "use_mpi" in control_flags
    known_orientations = "known_orientations" in control_flags

    # Only allow MPI tests if we've got MPI capabilities
    if use_mpi:
        pytest.importorskip("mpi4py")

    # Check the data files for this test exist
    sacla_path = dials_data("image_examples", pathlib=True)
    image_path = sacla_path / "SACLA-MPCCD-run266702-0-subset.h5"
    assert image_path.is_file()

    geometry_path = (
        sacla_path / "SACLA-MPCCD-run266702-0-subset-refined_experiments_level1.json"
    )
    assert geometry_path.is_file()

    # Write the .phil configuration to a file
    phil_path = tmp_path / "process_sacla.phil"
    with open(phil_path, "w") as f:
        f.write(sacla_phil % geometry_path)

        if known_orientations:
            known_orientations_path = os.path.join(
                sacla_path, "SACLA-MPCCD-run266702-0-subset-known_orientations.expt"
            )
            assert os.path.isfile(known_orientations_path)
            f.write("indexing.stills.known_orientations=%s\n" % known_orientations_path)
            f.write("indexing.stills.require_known_orientation=True\n")

    # Call dials.stills_process
    if use_mpi:
        command = [
            "mpirun",
            "-n",
            "4",
            "dials.stills_process",
            "mp.method=mpi mp.composite_stride=4 output.logging_dir=.",
        ]
    else:
        command = ["dials.stills_process"]
    command += [image_path, "process_sacla.phil"]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr

    def test_refl_table(result_filename, ranges):
        table = flex.reflection_table.from_file(result_filename)
        for expt_id, n_refls in enumerate(ranges):
            subset = table.select(table["id"] == expt_id)
            assert len(subset) in n_refls, (result_filename, expt_id, len(table))
        assert "id" in table
        assert set(table["id"]) == {0, 1, 2, 3}

    # large ranges to handle platform-specific differences
    test_refl_table(
        tmp_path / "idx-0000_integrated.refl",
        [
            list(range(140, 160)),
            list(range(575, 600)),
            list(range(420, 445)),
            list(range(485, 510)),
        ],
    )

    if known_orientations:
        test_refl_table(
            tmp_path / "idx-0000_coset6.refl",
            [
                list(range(155, 175)),
                list(range(545, 570)),
                list(range(430, 455)),
                list(range(480, 495)),
            ],
        )
    else:
        test_refl_table(
            tmp_path / "idx-0000_coset6.refl",
            [
                list(range(145, 160)),
                list(range(545, 570)),
                list(range(430, 455)),
                list(range(490, 515)),
            ],
        )


@pytest.mark.xfel
def test_pseudo_scan(dials_data, tmp_path):
    result = procrunner.run(
        (
            "dials.stills_process",
            dials_data("centroid_test_data", pathlib=True) / "centroid_000[1-2].cbf",
            "convert_sequences_to_stills=True",
            "squash_errors=False",
            "composite_output=True",
        ),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    experiments = ExperimentListFactory.from_json_file(
        tmp_path / "idx-0000_refined.expt", check_format=False
    )
    assert len(experiments) == 2



from __future__ import annotations

from dials_algorithms_image_threshold_ext import *  # noqa: F403; lgtm

__all__ = (  # noqa: F405
    "DispersionExtendedThreshold",
    "DispersionExtendedThresholdDebug",
    "DispersionThreshold",
    "DispersionThresholdDebug",
    "dispersion",
    "dispersion_w_gain",
    "gain",
    "index_of_dispersion",
    "index_of_dispersion_masked",
    "maximum_deviation",
    "niblack",
    "probability_distribution",
    "sauvola",
)



"""
Collection of factories for creating the scalers.
"""

from __future__ import annotations

import logging

from libtbx import Auto

from dials.algorithms.scaling.scaler import (
    MultiScaler,
    NullScaler,
    SingleScaler,
    TargetScaler,
)
from dials.algorithms.scaling.scaling_library import choose_initial_scaling_intensities
from dials.algorithms.scaling.scaling_utilities import (
    BadDatasetForScalingException,
    Reasons,
    align_axis_along_z,
    calc_crystal_frame_vectors,
    quasi_normalisation,
)
from dials.array_family import flex
from dials.util.filter_reflections import (
    filter_reflection_table_selection,
    sum_partial_reflections,
)

logger = logging.getLogger("dials")


def create_scaler(params, experiments, reflections):
    """Read an experimentlist and list of reflection tables and return
    an appropriate scaler. Requires experiment identifiers are correctly set in
    the experiments and reflections."""
    if not reflections:
        raise ValueError("No reflection tables provided as input")
    if len(reflections) == 1:
        scaler = SingleScalerFactory.create(params, experiments[0], reflections[0])
    else:
        is_scaled_list = [expt.scaling_model.is_scaled for expt in experiments]
        # if target datafile/model -> want to do targeted scaling only
        if params.scaling_options.reference:
            # last experiment/refl is target, rest are to scale against this
            scaler = TargetScalerFactory.create_for_target_against_reference(
                params, experiments, reflections
            )
        elif (  # not mtz/model, but some are scaled
            params.scaling_options.target_cycle and len(set(is_scaled_list)) == 2
        ):  # if only some scaled and want to do targeted scaling
            scaler = TargetScalerFactory.create(params, experiments, reflections)
        else:  # else just make one multiscaler for all refls
            scaler = MultiScalerFactory.create(params, experiments, reflections)
    return scaler


class ScalerFactory:
    """Base class for Scaler Factories"""

    @staticmethod
    def filter_bad_reflections(
        reflections, partiality_cutoff=0.4, min_isigi=-5.0, intensity_choice="combine"
    ):
        """Initial filter to select integrated reflections."""
        logger.info(
            "Applying filter of min_isigi > %s, partiality > %s",
            min_isigi,
            partiality_cutoff,
        )
        logger.disabled = True
        if intensity_choice == "combine":
            if "intensity.sum.value" not in reflections:
                if "intensity.prf.value" not in reflections:
                    intensity_choice = None
                else:
                    intensity_choice = ["profile"]
            elif "intensity.prf.value" not in reflections:
                intensity_choice = ["sum"]
            else:
                intensity_choice = ["sum | profile"]
        else:
            intensity_choice = [intensity_choice]
        if intensity_choice:
            good = filter_reflection_table_selection(
                reflections,
                intensity_choice=intensity_choice,
                combine_partials=False,
                partiality_threshold=partiality_cutoff,
                min_isigi=min_isigi,
            )
            mask = ~good
            reflections.set_flags(mask, reflections.flags.excluded_for_scaling)
        logger.disabled = False
        return reflections

    @staticmethod
    def ensure_experiment_identifier(experiment, reflection_table):
        """Check for consistent experiment identifier, and if not set then set it
        using scaled_id."""
        id_vals = list(reflection_table.experiment_identifiers().values())
        assert experiment.identifier in id_vals, (experiment.identifier, list(id_vals))
        assert len(id_vals) == 1, list(id_vals)
        logger.info(
            "The experiment id for this dataset is %s.",
            reflection_table.experiment_identifiers().keys()[0],
        )


class SingleScalerFactory(ScalerFactory):
    """Factory for creating a scaler for a single dataset."""

    @classmethod
    def create(cls, params, experiment, reflection_table, for_multi=False):
        """Perform reflection_table preprocessing and create a SingleScaler."""

        cls.ensure_experiment_identifier(experiment, reflection_table)

        logger.info(
            "The scaling model type being applied is %s. \n",
            experiment.scaling_model.id_,
        )
        try:
            reflection_table = cls.filter_bad_reflections(
                reflection_table,
                partiality_cutoff=params.cut_data.partiality_cutoff,
                min_isigi=params.cut_data.min_isigi,
                intensity_choice=params.reflection_selection.intensity_choice,
            )
        except ValueError:
            raise BadDatasetForScalingException

        # combine partial measurements of same reflection, to handle those reflections
        # that were split by dials.integrate  - changes size of reflection table.
        reflection_table = sum_partial_reflections(reflection_table)

        if "inverse_scale_factor" not in reflection_table:
            reflection_table["inverse_scale_factor"] = flex.double(
                reflection_table.size(), 1.0
            )
        elif (
            reflection_table["inverse_scale_factor"].count(0.0)
            == reflection_table.size()
        ):
            reflection_table["inverse_scale_factor"] = flex.double(
                reflection_table.size(), 1.0
            )
        reflection_table = choose_initial_scaling_intensities(
            reflection_table, params.reflection_selection.intensity_choice
        )

        excluded_for_scaling = reflection_table.get_flags(
            reflection_table.flags.excluded_for_scaling
        )
        user_excluded = reflection_table.get_flags(
            reflection_table.flags.user_excluded_in_scaling
        )
        reasons = Reasons()
        reasons.add_reason("user excluded", user_excluded.count(True))
        reasons.add_reason("excluded for scaling", excluded_for_scaling.count(True))
        n_excluded = (excluded_for_scaling | user_excluded).count(True)
        if n_excluded == reflection_table.size():
            logger.info("All reflections were determined to be unsuitable for scaling.")
            logger.info(reasons)
            raise BadDatasetForScalingException(
                """Unable to use this dataset for scaling"""
            )
        else:
            logger.info(
                "Excluding %s/%s reflections\n%s",
                n_excluded,
                reflection_table.size(),
                reasons,
            )

        if params.reflection_selection.method == "intensity_ranges":
            reflection_table = quasi_normalisation(reflection_table, experiment)
        if (
            params.reflection_selection.method in (None, Auto, "auto", "quasi_random")
        ) or (
            experiment.scaling_model.id_ == "physical"
            and "absorption" in experiment.scaling_model.components
        ):
            if experiment.scan:
                reflection_table = calc_crystal_frame_vectors(
                    reflection_table, experiment
                )
                alignment_axis = (1.0, 0.0, 0.0)
                reflection_table["s0c"] = align_axis_along_z(
                    alignment_axis, reflection_table["s0c"]
                )
                reflection_table["s1c"] = align_axis_along_z(
                    alignment_axis, reflection_table["s1c"]
                )
        try:
            scaler = SingleScaler(params, experiment, reflection_table, for_multi)
        except BadDatasetForScalingException as e:
            raise ValueError(e)
        else:
            return scaler


class NullScalerFactory(ScalerFactory):
    "Factory for creating null scaler"

    @classmethod
    def create(cls, params, experiment, reflection_table):
        """Return Null Scaler."""

        logger.info("Preprocessing target dataset for scaling. \n")
        reflection_table = cls.filter_bad_reflections(reflection_table)
        if "variance" in reflection_table:
            variance_mask = reflection_table["variance"] <= 0.0
            reflection_table.set_flags(
                variance_mask, reflection_table.flags.excluded_for_scaling
            )
        else:
            reflection_table.set_flags(
                flex.bool(reflection_table.size(), False),
                reflection_table.flags.excluded_for_scaling,
            )
        logger.info(
            "%s reflections not suitable for scaling\n",
            reflection_table.get_flags(
                reflection_table.flags.excluded_for_scaling
            ).count(True),
        )
        cls.ensure_experiment_identifier(experiment, reflection_table)
        return NullScaler(params, experiment, reflection_table)


class MultiScalerFactory:
    "Factory for creating a scaler for multiple datasets"

    @staticmethod
    def create(params, experiments, reflections):
        """create a list of single scalers to pass to a MultiScaler."""
        single_scalers = []
        idx_to_remove = []
        for i, (expt, refl) in enumerate(zip(experiments, reflections)):
            # Remove bad datasets that literally have no integrated reflections
            try:
                scaler = SingleScalerFactory.create(params, expt, refl, for_multi=True)
            except BadDatasetForScalingException as e:
                logger.info(e)
                idx_to_remove.append(i)
            else:
                single_scalers.append(scaler)
        ms = MultiScaler(single_scalers)
        if idx_to_remove:
            for i in idx_to_remove:
                ms.removed_datasets.append(experiments[i].identifier)
            logger.info(
                "Removed experiments %s", " ".join(str(i) for i in idx_to_remove)
            )
        return ms

    @staticmethod
    def create_from_targetscaler(targetscaler):
        """method to pass scalers from TargetScaler to a MultiScaler"""
        single_scalers = []
        for scaler in targetscaler.unscaled_scalers:
            single_scalers.append(scaler)
        single_scalers.extend(targetscaler.single_scalers)
        multiscaler = MultiScaler(single_scalers)
        multiscaler.observers = targetscaler.observers
        return multiscaler


class TargetScalerFactory:
    "Factory for creating a targeted scaler for multiple datasets"

    @staticmethod
    def create_for_target_against_reference(params, experiments, reflections):
        """Create TargetScaler for case where have a reference datafile or model."""
        scaled_scalers = []
        unscaled_scalers = []
        idx_to_remove = []

        for i, (expt, refl) in enumerate(zip(experiments[:-1], reflections[:-1])):
            # Remove bad datasets that literally have no integrated reflections
            try:
                scaler = SingleScalerFactory.create(params, expt, refl, for_multi=True)
            except BadDatasetForScalingException as e:
                logger.info(e)
                idx_to_remove.append(i)
            else:
                unscaled_scalers.append(scaler)
        scaled_scalers = [
            NullScalerFactory.create(params, experiments[-1], reflections[-1])
        ]
        ts = TargetScaler(scaled_scalers, unscaled_scalers)
        if idx_to_remove:
            for i in idx_to_remove:
                ts.removed_datasets.append(experiments[i].identifier)
            logger.info(
                "Removed experiments %s", " ".join(str(i) for i in idx_to_remove)
            )
        return ts

    @staticmethod
    def create(params, experiments, reflections):
        """sort scaled and unscaled datasets to pass to TargetScaler"""
        scaled_experiments = []
        scaled_scalers = []
        unscaled_scalers = []
        idx_to_remove = []

        for i, (expt, refl) in enumerate(zip(experiments, reflections)):
            # Remove bad datasets that literally have no integrated reflections
            try:
                scaler = SingleScalerFactory.create(params, expt, refl, for_multi=True)
            except BadDatasetForScalingException as e:
                logger.info(e)
                idx_to_remove.append(i)
            else:
                if expt.scaling_model.is_scaled:
                    scaled_scalers.append(scaler)
                    scaled_experiments.append(expt)
                else:
                    unscaled_scalers.append(scaler)
        if idx_to_remove:
            for j in idx_to_remove[::-1]:
                del experiments[j]
                del reflections[j]
            logger.info(
                "Removed experiments %s", " ".join(str(i) for i in idx_to_remove)
            )

        n_exp, n_refl = (len(experiments), len(reflections))
        n_ss, n_us = (len(scaled_scalers), len(unscaled_scalers))
        assert n_exp == n_ss + n_us, (n_exp, str(n_ss) + " + " + str(n_us))
        assert n_exp == n_refl, (n_exp, n_refl)
        return TargetScaler(scaled_scalers, unscaled_scalers)



from __future__ import annotations

import logging
import math
import os
import re

from scitbx import matrix

from dials.util.filter_reflections import filter_reflection_table

logger = logging.getLogger(__name__)


def export_sadabs(integrated_data, experiment_list, params):
    """Export data from integrated_data corresponding to experiment_list to a
    file for input to SADABS. FIXME probably need to make a .p4p file as
    well..."""

    from dials.array_family import flex

    # for the moment assume (and assert) that we will convert data from exactly
    # one lattice...

    assert len(experiment_list) == 1
    # select reflections that are assigned to an experiment (i.e. non-negative id)

    integrated_data = integrated_data.select(integrated_data["id"] >= 0)
    assert max(integrated_data["id"]) == 0

    # export for sadabs should only be for non-scaled reflections
    assert any(
        i in integrated_data for i in ["intensity.sum.value", "intensity.prf.value"]
    )

    integrated_data = filter_reflection_table(
        integrated_data,
        intensity_choice=params.intensity,
        partiality_threshold=params.mtz.partiality_threshold,
        combine_partials=params.mtz.combine_partials,
        min_isigi=params.mtz.min_isigi,
        filter_ice_rings=params.mtz.filter_ice_rings,
        d_min=params.mtz.d_min,
    )

    experiment = experiment_list[0]
    assert experiment.scan is not None

    # sort data before output
    nref = len(integrated_data["miller_index"])
    indices = flex.size_t_range(nref)
    perm = sorted(indices, key=lambda k: integrated_data["miller_index"][k])
    integrated_data = integrated_data.select(flex.size_t(perm))

    assert experiment.goniometer is not None

    # Warn of unhelpful SADABS behaviour for certain multi-sequence data sets
    hkl_file_root, _ = os.path.splitext(params.sadabs.hklout)
    if not params.sadabs.run or re.search("_0+$", hkl_file_root):
        logger.warning(
            "It seems SADABS rejects multi-sequence data when the first "
            "filename ends "
            "'_0', '_00', etc., with a cryptic error message:\n"
            "\t'Inconsistent 2theta values in same scan'.\n"
            "You may need to begin the numbering of your SADABS HKL files from 1, "
            "rather than 0, and ensure the SADABS run/batch number is greater than 0."
        )

    axis = matrix.col(experiment.goniometer.get_rotation_axis_datum())

    beam = matrix.col(experiment.beam.get_sample_to_source_direction())
    s0 = matrix.col(experiment.beam.get_s0())

    F = matrix.sqr(experiment.goniometer.get_fixed_rotation())
    S = matrix.sqr(experiment.goniometer.get_setting_rotation())
    unit_cell = experiment.crystal.get_unit_cell()

    if params.debug:
        m_format = "%6.3f%6.3f%6.3f\n%6.3f%6.3f%6.3f\n%6.3f%6.3f%6.3f"
        c_format = "%.2f %.2f %.2f %.2f %.2f %.2f"

        logger.info(
            "Unit cell parameters from experiment: %s",
            c_format % unit_cell.parameters(),
        )
        logger.info(
            "Symmetry: %s", experiment.crystal.get_space_group().type().lookup_symbol()
        )

        logger.info("Goniometer fixed matrix:\n%s", m_format % F.elems)
        logger.info("Goniometer setting matrix:\n%s", m_format % S.elems)
        logger.info("Goniometer scan axis:\n%6.3f%6.3f%6.3f", axis.elems)

    # detector scaling info
    assert len(experiment.detector) == 1
    panel = experiment.detector[0]
    dims = panel.get_image_size()
    pixel = panel.get_pixel_size()
    fast_axis = matrix.col(panel.get_fast_axis())
    slow_axis = matrix.col(panel.get_slow_axis())
    normal = fast_axis.cross(slow_axis)
    detector2t = s0.angle(normal, deg=True)

    if params.debug:
        logger.info("Detector fast, slow axes:")
        logger.info("%6.3f%6.3f%6.3f", fast_axis.elems)
        logger.info("%6.3f%6.3f%6.3f", slow_axis.elems)
        logger.info("Detector two theta (degrees): %.2f", detector2t)

    scl_x = 512.0 / (dims[0] * pixel[0])
    scl_y = 512.0 / (dims[1] * pixel[1])

    image_range = experiment.scan.get_image_range()

    from cctbx.array_family import flex as cflex  # implicit import # noqa: F401
    from cctbx.miller import map_to_asu_isym  # implicit import # noqa: F401

    # gather the required information for the reflection file

    nref = len(integrated_data["miller_index"])

    miller_index = integrated_data["miller_index"]

    if "intensity.sum.value" in integrated_data:
        I = integrated_data["intensity.sum.value"]
        V = integrated_data["intensity.sum.variance"]
        assert V.all_gt(0)
        sigI = flex.sqrt(V)
    else:
        I = integrated_data["intensity.prf.value"]
        V = integrated_data["intensity.prf.variance"]
        assert V.all_gt(0)
        sigI = flex.sqrt(V)

    # figure out scaling to make sure data fit into format 2F8.2 i.e. Imax < 1e5

    Imax = flex.max(I)

    if params.debug:
        logger.info("Maximum intensity in file: %8.2f", Imax)

    if Imax > 99999.0:
        scale = 99999.0 / Imax
        I = I * scale
        sigI = sigI * scale

    phi_start, phi_range = experiment.scan.get_image_oscillation(image_range[0])

    if params.sadabs.predict:
        logger.info("Using scan static predicted spot locations")
        from dials.algorithms.spot_prediction import ScanStaticReflectionPredictor

        predictor = ScanStaticReflectionPredictor(experiment)
        UB = experiment.crystal.get_A()
        predictor.for_reflection_table(integrated_data, UB)

    if not experiment.crystal.num_scan_points:
        logger.info("No scan varying model: use static")
        static = True
    else:
        static = False

    with open(params.sadabs.hklout, "w") as fout:

        for j in range(nref):

            h, k, l = miller_index[j]

            if params.sadabs.predict:
                x_mm, y_mm, z_rad = integrated_data["xyzcal.mm"][j]
            else:
                x_mm, y_mm, z_rad = integrated_data["xyzobs.mm.value"][j]

            z0 = integrated_data["xyzcal.px"][j][2]
            istol = int(round(10000 * unit_cell.stol((h, k, l))))

            if params.sadabs.predict or static:
                # work from a scan static model & assume perfect goniometer
                # FIXME maybe should work back in the option to predict spot positions
                UB = matrix.sqr(experiment.crystal.get_A())
                phi = phi_start + z0 * phi_range
                R = axis.axis_and_angle_as_r3_rotation_matrix(phi, deg=True)
                RUB = S * R * F * UB
            else:
                # properly compute RUB for every reflection
                UB = matrix.sqr(experiment.crystal.get_A_at_scan_point(int(round(z0))))
                phi = phi_start + z0 * phi_range
                R = axis.axis_and_angle_as_r3_rotation_matrix(phi, deg=True)
                RUB = S * R * F * UB

            x = RUB * (h, k, l)
            s = (s0 + x).normalize()

            # can also compute s based on centre of mass of spot
            # s = (origin + x_mm * fast_axis + y_mm * slow_axis).normalize()

            astar = (RUB * (1, 0, 0)).normalize()
            bstar = (RUB * (0, 1, 0)).normalize()
            cstar = (RUB * (0, 0, 1)).normalize()

            ix = beam.dot(astar)
            iy = beam.dot(bstar)
            iz = beam.dot(cstar)

            dx = s.dot(astar)
            dy = s.dot(bstar)
            dz = s.dot(cstar)

            x = x_mm * scl_x
            y = y_mm * scl_y
            z = (z_rad * 180 / math.pi - phi_start) / phi_range

            fout.write(
                "%4d%4d%4d%8.2f%8.2f%4d%8.5f%8.5f%8.5f%8.5f%8.5f%8.5f"
                % (h, k, l, I[j], sigI[j], params.sadabs.run, ix, dx, iy, dy, iz, dz)
            )
            fout.write("%7.2f%7.2f%8.2f%7.2f%5d\n" % (x, y, z, detector2t, istol))

    fout.close()
    logger.info("Output %d reflections to %s", nref, params.sadabs.hklout)



from __future__ import annotations

import logging

logger = logging.getLogger(__name__)

from rstbx.symmetry.constraints.parameter_reduction import symmetrize_reduce_enlarge
from scitbx import matrix

from dials.algorithms.refinement.parameterisation.model_parameters import (
    ModelParameterisation,
    Parameter,
)
from dials.algorithms.refinement.refinement_helpers import CrystalOrientationCompose


class CrystalOrientationMixin:
    """Mix-in class defining some functionality unique to crystal orientation
    parameterisations that can be shared by static and scan-varying versions"""

    @staticmethod
    def _build_p_list(parameter_type=Parameter):
        """Build the list of parameters, using the parameter_type callback to
        select between versions of the Parameter class"""

        # set up the parameters
        phi1 = parameter_type(0.0, matrix.col((1, 0, 0)), "angle (mrad)", "Phi1")
        phi2 = parameter_type(0.0, matrix.col((0, 1, 0)), "angle (mrad)", "Phi2")
        phi3 = parameter_type(0.0, matrix.col((0, 0, 1)), "angle (mrad)", "Phi3")

        # build the parameter list in a specific,  maintained order
        p_list = [phi1, phi2, phi3]

        return p_list


class CrystalOrientationParameterisation(
    ModelParameterisation, CrystalOrientationMixin
):
    """A parameterisation of the orientation of a Crystal model.

    The Crystal orientation matrix U is parameterised by three Tait-Bryan angles
    expressed in mrad"""

    def __init__(self, crystal, experiment_ids=None):
        """Initialise the CrystalOrientationParameterisation object

        Args:
            crystal: A dxtbx Crystal object to be parameterised.
            experiment_ids (list): The experiment IDs affected by this
                parameterisation. Defaults to None, which is replaced by [0].
        """

        # The state of a crystal orientation parameterisation is an orientation
        # matrix '[U]'. The initial state is a snapshot of the crystal
        # orientation at the time of initialisation '[U0]'. Future states are
        # composed by rotations around axes of the phi-axis frame by Tait-Bryan
        # angles.
        #
        # [U] = [Phi3][Phi2][Phi1][U0]

        # set up the initial state
        if experiment_ids is None:
            experiment_ids = [0]
        istate = matrix.sqr(crystal.get_U())

        # build the parameter list
        p_list = self._build_p_list()

        # set up the base class
        ModelParameterisation.__init__(
            self, crystal, istate, p_list, experiment_ids=experiment_ids
        )

        # call compose to calculate all the derivatives
        self.compose()

        return

    def compose(self):

        # Extract orientation from the initial state
        U0 = self._initial_state

        # extract parameters from the internal list
        phi1, phi2, phi3 = self._param

        # calculate using the helper class
        coc = CrystalOrientationCompose(
            U0, phi1.value, phi1.axis, phi2.value, phi2.axis, phi3.value, phi3.axis
        )

        # compose new state
        self._model.set_U(coc.U())

        # store derivatives
        self._dstate_dp = [coc.dU_dphi1(), coc.dU_dphi2(), coc.dU_dphi3()]

        return

    def get_state(self):

        # only a single crystal is parameterised here, so no multi_state_elt
        # argument is allowed
        return matrix.sqr(self._model.get_U())


class CrystalUnitCellMixin:
    """Mix-in class defining some functionality unique to crystal unit cell
    parameterisations that can be shared by static and scan-varying versions"""

    def _build_p_list(self, crystal, parameter_type=Parameter):
        """Build the list of parameters, using the parameter_type callback to
        select between versions of the Parameter class"""

        # Set up symmetrizing object
        S = symmetrize_reduce_enlarge(crystal.get_space_group())
        S.set_orientation(orientation=crystal.get_B())
        X = S.forward_independent_parameters()

        # Set up the independent parameters, with a change of scale
        p_list = [
            parameter_type(e * 1.0e5, name="g_param_%d" % i) for i, e in enumerate(X)
        ]

        return p_list

    def _compose_core(self, raw_vals):

        # obtain metrical matrix parameters on natural scale
        vals = [v * 1.0e-5 for v in raw_vals]

        # set parameter values in the symmetrizing object and obtain new B
        S = symmetrize_reduce_enlarge(self._model.get_space_group())
        S.set_orientation(orientation=self._model.get_B())
        S.forward_independent_parameters()  # Set Bconverter as side-effect
        try:
            newB = matrix.sqr(S.backward_orientation(vals).reciprocal_matrix())
        except RuntimeError as e:

            # write original error to debug log
            logger.debug("Unable to compose the crystal model")
            logger.debug("Original error message: %s", str(e))
            logger.debug("Failing now.")
            raise RuntimeError(
                "Unable to compose the crystal model. Please check that the "
                "experiments match the indexing of the reflections."
            )

        # returns the independent parameters given the set_orientation() B matrix
        # used here for side effects
        S.forward_independent_parameters()

        # get the derivatives of state wrt metrical matrix parameters on the
        # adjusted sale
        dB_dval = [matrix.sqr(g) * 1.0e-5 for g in S.forward_gradients()]

        return newB, dB_dval


class CrystalUnitCellParameterisation(ModelParameterisation, CrystalUnitCellMixin):
    """A parameterisation for the unit cell of a Crystal model.

    The Crystal reciprocal space orthogonalisation matrix B is parameterised
    using up to 6 metrical matrix elements, rescaled by a multiplicative factor.
    """

    def __init__(self, crystal, experiment_ids=None):
        """Initialise the CrystalUnitCellParameterisation object

        Args:
            crystal: A dxtbx Crystal object to be parameterised.
            experiment_ids (list): The experiment IDs affected by this
                parameterisation. Defaults to None, which is replaced by [0].
        """

        # The state of the unit cell parameterisation is the reciprocal space
        # orthogonalisation matrix 'B'. The initial state is irrelevant for
        # this model, as composition of a new B matrix and derivatives can be
        # done with just the values of 6 unit cell parameters, without
        # defining axial directions (which are selected by choice of the PDB
        # convention). For this reason also, the axes of the
        # parameters are irrelevant and are set here to None.

        ### Set up the initial state
        if experiment_ids is None:
            experiment_ids = [0]
        istate = None

        # build the parameter list
        p_list = self._build_p_list(crystal)

        # set up the base class
        ModelParameterisation.__init__(
            self, crystal, istate, p_list, experiment_ids=experiment_ids
        )

        # call compose to calculate all the derivatives
        self.compose()

        return

    def compose(self):

        # calculate new B and derivatives
        newB, self._dstate_dp = self._compose_core([p.value for p in self._param])

        # Now pass new B to the crystal model
        self._model.set_B(newB)

        return

    def get_state(self):

        # only a single crystal is parameterised here, so no multi_state_elt
        # argument is allowed
        return matrix.sqr(self._model.get_B())

    def set_state_uncertainties(self, var_cov, multi_state_elt=None):

        self._model.set_B_covariance(var_cov)

        return



from __future__ import annotations

import collections
import math
from copy import deepcopy

import numpy as np

from dxtbx.format.nexus import h5str
from scitbx import matrix

# Extensions to NXMX
#
#  "detector/underload" - trusted_range[0]
#  "detector/timestamp" - epochs
#  "entry/index" - experiment index and shared model indices
#  "entry/template" - save location to external files (e.g. cbf)

schema_url = (
    "https://github.com/nexusformat/definitions/blob/master/applications/NXmx.nxdl.xml"
)


def convert_to_nexus_beam_direction(experiments):
    EPS = 1e-7

    zaxis = matrix.col((0, 0, -1))

    class Dummy:
        pass

    experiments2 = []

    # Save the sharedness as the id
    for exp in experiments:
        exp.beam_id = id(exp.beam)
        exp.detector_id = id(exp.detector)
        exp.goniometer_id = id(exp.goniometer)
        exp.scan_id = id(exp.scan)
        exp.crystal_id = id(exp.crystal)
        exp2 = Dummy()
        exp2.beam_id = id(exp.beam)
        exp2.detector_id = id(exp.detector)
        exp2.goniometer_id = id(exp.goniometer)
        exp2.scan_id = id(exp.scan)
        exp2.crystal_id = id(exp.crystal)
        experiments2.append(exp2)

    # Copy all models
    for exp, exp2 in zip(experiments, experiments2):
        exp2.beam = deepcopy(exp.beam)
        exp2.detector = deepcopy(exp.detector)
        exp2.goniometer = deepcopy(exp.goniometer)
        exp2.scan = deepcopy(exp.scan)
        exp2.crystal = deepcopy(exp.crystal)
        exp2.imageset = exp.imageset
    experiments = experiments2

    # Rotate the beams
    rotations = []
    for exp in experiments:
        d = matrix.col(exp.beam.get_sample_to_source_direction()).normalize()
        angle = d.angle(zaxis, deg=False)
        if abs(angle - math.pi) < EPS:
            axis = (1, 0, 0)
        elif abs(angle) < EPS:
            rotations.append(((0, 0, 0), 0))
            continue
        else:
            axis = d.cross(zaxis).normalize()
        exp.beam.rotate_around_origin(axis, angle, deg=False)
        exp.detector.rotate_around_origin(axis, angle, deg=False)
        if exp.goniometer:
            exp.goniometer.rotate_around_origin(axis, angle, deg=False)
        exp.crystal.rotate_around_origin(axis, angle, deg=False)
        d = matrix.col(exp.beam.get_sample_to_source_direction())
        assert abs(d.angle(zaxis)) < EPS
        rotations.append((axis, angle))

    # Return converted experiments
    return experiments, rotations


def convert_from_nexus_beam_direction(experiments, rotations):
    EPS = 1e-7

    zaxis = matrix.col((0, 0, -1))
    for exp, (axis, angle) in zip(experiments, rotations):
        d = matrix.col(exp.beam.get_sample_to_source_direction()).normalize()
        assert abs(d.angle(zaxis)) < EPS
        exp.beam.rotate_around_origin(axis, angle, deg=False)
        exp.detector.rotate_around_origin(axis, angle, deg=False)
        if exp.goniometer:
            exp.goniometer.rotate_around_origin(axis, angle, deg=False)
        exp.crystal.rotate_around_origin(axis, angle, deg=False)
    return experiments


def polarization_normal_to_stokes(n, p):
    EPS = 1e-7
    ax = matrix.col((1, 0, 0))
    ay = matrix.col((0, 1, 0))
    az = matrix.col((0, 0, 1))
    n = matrix.col(n).normalize()
    assert abs(n.dot(az)) < EPS
    I = 1.0
    X = 0.0
    W = math.atan2(n.dot(ay), n.dot(ax)) - math.pi / 2
    S0 = I
    S1 = I * p * math.cos(2 * W) * math.cos(2 * X)
    S2 = I * p * math.sin(2 * W) * math.cos(2 * X)
    S3 = I * p * math.sin(2 * X)
    return S0, S1, S2, S3


def polarization_stokes_to_normal(S0, S1, S2, S3):
    EPS = 1e-7
    p = math.sqrt(S1 * S1 + S2 * S2 + S3 * S3) / S0
    W = math.atan2(S2, S1) * 0.5
    X = math.atan2(S3, math.sqrt(S1 * S1 + S2 * S2)) * 0.5
    assert abs(X) < EPS
    n = matrix.col((-math.sin(W), math.cos(W), 0.0)).normalize()
    return n, p


def get_nx_class(handle, nx_class, path):
    if path in handle:
        group = handle[path]
        assert group.attrs["NX_class"] == nx_class
    else:
        group = handle.create_group(path)
        group.attrs["NX_class"] = nx_class
    return group


def get_nx_sample(handle, path):
    return get_nx_class(handle, "NXsample", path)


def get_nx_beam(handle, path):
    return get_nx_class(handle, "NXbeam", path)


def get_nx_detector(handle, path):
    return get_nx_class(handle, "NXdetector", path)


def get_nx_detector_module(handle, path):
    return get_nx_class(handle, "NXdetector_module", path)


def get_nx_instrument(handle, path):
    return get_nx_class(handle, "NXinstrument", path)


def get_nx_transformations(handle, path):
    return get_nx_class(handle, "NXtransformations", path)


def get_nx_process(handle, path):
    return get_nx_class(handle, "NXprocess", path)


def get_nx_note(handle, path):
    return get_nx_class(handle, "NXnote", path)


def get_nx_source(handle, path):
    return get_nx_class(handle, "NXsource", path)


def get_nx_data(handle, path, data):
    handle[path] = data
    return handle[path]


def get_nx_dials(handle, path):
    if path in handle:
        group = handle[path]
        if "NX_class" in group.attrs:
            # Backwards compatibility. See https://github.com/dials/dials/issues/1124
            assert group.attrs["NX_class"] == "NXdials"
    else:
        group = handle.create_group(path)
    return group


def dump_beam(entry, beam):
    """Export the beam model."""
    EPS = 1e-7

    # Get the nx_beam
    nx_sample = get_nx_sample(entry, "sample")
    nx_beam = get_nx_beam(nx_sample, "beam")

    # Generate the stokes polarization parameters
    d = matrix.col(beam.get_sample_to_source_direction()).normalize()
    n = matrix.col(beam.get_polarization_normal()).normalize()
    p = beam.get_polarization_fraction()
    assert abs(n.dot(d)) < EPS
    assert abs(d.dot(matrix.col((0, 0, -1))) - 1) < EPS

    # Get the polarization in stokes parameters
    S0, S1, S2, S3 = polarization_normal_to_stokes(n, p)

    # Set the beam parameters
    nx_beam["incident_wavelength"] = beam.get_wavelength()
    nx_beam["incident_wavelength"].attrs["units"] = "angstrom"
    nx_beam["incident_polarization_stokes"] = (S0, S1, S2, S3)


def dump_detector(entry, detector, beam, imageset, scan):
    EPS = 1e-7

    # Get the detector
    nx_instrument = get_nx_instrument(entry, "instrument")
    nx_detector = get_nx_detector(nx_instrument, "detector")

    # Get some bulk properties
    thickness = detector[0].get_thickness()
    material = detector[0].get_material()
    dtype = detector[0].get_type()
    trusted_range = detector[0].get_trusted_range()

    # Check all panels obey these bulk properties
    assert [abs((p.get_thickness() - thickness) < EPS) for p in detector].count(
        False
    ) == 0
    assert [p.get_material() == material for p in detector].count(False) == 0
    assert [p.get_type() == dtype for p in detector].count(False) == 0
    assert [p.get_trusted_range() == trusted_range for p in detector].count(False) == 0

    # Take the distance and beam centre from the first panel
    # distance = detector[0].get_directed_distance()
    # beam_centre_x, beam_centre_y = detector[0].get_beam_centre(beam.get_s0())

    # Set the detector properties
    nx_detector["sensor_thickness"] = thickness
    nx_detector["sensor_thickness"].attrs["units"] = "mm"
    nx_detector["sensor_material"] = material
    nx_detector["type"] = dtype
    # nx_detector['distance'] = distance
    # nx_detector['beam_centre_x'] = beam_centre_x
    # nx_detector['beam_centre_y'] = beam_centre_y
    nx_detector["saturation_value"] = int(trusted_range[1])
    nx_detector["underload"] = int(trusted_range[0])  # FIXME non-standard
    nx_detector["description"] = dtype

    # Make up some fake stuff
    if scan is not None:
        nx_detector[
            "timestamp"
        ] = scan.get_epochs().as_numpy_array()  # FIXME non-standard
        nx_detector["frame_time"] = scan.get_exposure_times().as_numpy_array()
        nx_detector["frame_time"].attrs["units"] = "s"
        # Optional so don't try
        # nx_detector["dead_time"] = [0.0] * len(scan)
        # nx_detector["count_time"] = [0.0] * len(scan)
        # nx_detector["detector_readout_time"] = [0.0] * len(scan)
    else:
        # Optional so don't try
        pass
        # nx_detector["timestamp"] = 0.0
        # nx_detector["dead_time"] = 0.0
        # nx_detector["count_time"] = 0.0
        # nx_detector["frame_time"] = 0.0
        # nx_detector["detector_readout_time"] = 0.0
    nx_detector["bit_depth_readout"] = 32

    # Create the detector depends on
    nx_detector["depends_on"] = "."

    # Create some data for example file
    # data = [im.as_numpy_array() for im in imageset]

    # Create the nx data
    # nx_data = get_nx_data(nx_detector, "data", [[[0]],[[0]],[[0]]])

    # Loop through all the panels
    for i, panel in enumerate(detector):

        # Get some panel attributes
        pixel_size = panel.get_pixel_size()
        image_size = panel.get_image_size()
        origin = matrix.col(panel.get_origin())

        # Get the detector module object
        nx_module = get_nx_detector_module(nx_detector, "module%d" % i)

        # Set the data size
        nx_module["data_size"] = image_size
        nx_module["data_origin"] = (-1, -1)  # FIXME INVALID

        # Set the module offset
        nx_module["module_offset"] = origin.length()
        nx_module["module_offset"].attrs["depends_on"] = "."
        nx_module["module_offset"].attrs["transformation_type"] = "translation"
        nx_module["module_offset"].attrs["offset"] = (0.0, 0.0, 0.0)
        nx_module["module_offset"].attrs["vector"] = origin.normalize()
        nx_module["module_offset"].attrs["units"] = "mm"

        # The path for items below
        module_offset_path = str(nx_module["module_offset"].name)

        # Write the fast pixel direction
        nx_module["fast_pixel_direction"] = pixel_size[0]
        nx_module["fast_pixel_direction"].attrs["depends_on"] = module_offset_path
        nx_module["fast_pixel_direction"].attrs["transformation_type"] = "translation"
        nx_module["fast_pixel_direction"].attrs["offset"] = (0.0, 0.0, 0.0)
        nx_module["fast_pixel_direction"].attrs["vector"] = panel.get_fast_axis()
        nx_module["fast_pixel_direction"].attrs["units"] = "mm"

        # Write the slow pixel direction
        nx_module["slow_pixel_direction"] = pixel_size[1]
        nx_module["slow_pixel_direction"].attrs["depends_on"] = module_offset_path
        nx_module["slow_pixel_direction"].attrs["transformation_type"] = "translation"
        nx_module["slow_pixel_direction"].attrs["offset"] = (0.0, 0.0, 0.0)
        nx_module["slow_pixel_direction"].attrs["vector"] = panel.get_slow_axis()
        nx_module["slow_pixel_direction"].attrs["units"] = "mm"


def dump_goniometer(entry, goniometer, scan):
    """Export the goniometer model."""
    if scan is None or goniometer is None:
        return

    # The angles for each image
    phi0, dphi = scan.get_oscillation(deg=True)
    phi = [phi0 + dphi * i for i in range(len(scan))]

    nx_sample = get_nx_sample(entry, "sample")
    nx_transformations = get_nx_transformations(nx_sample, "transformations")

    # Write out the rotation axis and oscillation
    r = matrix.sqr(goniometer.get_setting_rotation())
    q = r.r3_rotation_matrix_as_unit_quaternion()
    angle, axis = q.unit_quaternion_as_axis_and_angle()
    nx_transformations["setting_rotation"] = angle
    nx_transformations["setting_rotation"].attrs["depends_on"] = "."
    nx_transformations["setting_rotation"].attrs["transformation_type"] = "rotation"
    nx_transformations["setting_rotation"].attrs["offset_units"] = "mm"
    nx_transformations["setting_rotation"].attrs["offset"] = (0.0, 0.0, 0.0)
    nx_transformations["setting_rotation"].attrs["vector"] = axis

    r = matrix.sqr(goniometer.get_fixed_rotation())
    q = r.r3_rotation_matrix_as_unit_quaternion()
    angle, axis = q.unit_quaternion_as_axis_and_angle()
    nx_transformations["fixed_rotation"] = angle
    nx_transformations["fixed_rotation"].attrs["depends_on"] = str(
        nx_transformations["setting_rotation"].name
    )
    nx_transformations["fixed_rotation"].attrs["transformation_type"] = "rotation"
    nx_transformations["fixed_rotation"].attrs["offset_units"] = "mm"
    nx_transformations["fixed_rotation"].attrs["offset"] = (0.0, 0.0, 0.0)
    nx_transformations["fixed_rotation"].attrs["vector"] = axis

    nx_transformations["phi"] = phi
    nx_transformations["phi"].attrs["depends_on"] = str(
        nx_transformations["fixed_rotation"].name
    )
    nx_transformations["phi"].attrs["transformation_type"] = "rotation"
    nx_transformations["phi"].attrs["offset_units"] = "mm"
    nx_transformations["phi"].attrs["offset"] = (0.0, 0.0, 0.0)
    nx_transformations["phi"].attrs["vector"] = goniometer.get_rotation_axis_datum()


def dump_crystal(entry, crystal, scan):
    """Export the crystal model."""
    from scitbx.array_family import flex

    # Get the sample
    nx_sample = get_nx_sample(entry, "sample")

    # Set the space group
    nx_sample["unit_cell_group"] = crystal.get_space_group().type().hall_symbol()

    # Get the unit cell and orientation matrix in the case of scan varying and
    # scan static models
    if crystal.num_scan_points:
        num = crystal.num_scan_points
        unit_cell = flex.double(flex.grid(num, 6))
        orientation_matrix = flex.double(flex.grid(num, 9))
        for i in range(num):
            __cell = crystal.get_unit_cell_at_scan_point(i).parameters()
            for j in range(6):
                unit_cell[i, j] = __cell[j]
            __matrix = crystal.get_U_at_scan_point(i)
            for j in range(9):
                orientation_matrix[i, j] = __matrix[j]
        orientation_matrix = [
            [
                tuple(orientation_matrix[i : i + 1, 0:3]),
                tuple(orientation_matrix[i : i + 1, 3:6]),
                tuple(orientation_matrix[i : i + 1, 6:9]),
            ]
            for i in range(num)
        ]
        unit_cell = [tuple(unit_cell[i : i + 1, :]) for i in range(num)]
        average_unit_cell = crystal.get_unit_cell().parameters()
        average_orientation_matrix = crystal.get_U()
        average_orientation_matrix = [
            average_orientation_matrix[0:3],
            average_orientation_matrix[3:6],
            average_orientation_matrix[6:9],
        ]

    else:
        unit_cell = [crystal.get_unit_cell().parameters()]
        orientation_matrix = [crystal.get_U()]
        orientation_matrix = [
            [
                tuple(orientation_matrix[0][0:3]),
                tuple(orientation_matrix[0][3:6]),
                tuple(orientation_matrix[0][6:9]),
            ]
        ]
        average_unit_cell = unit_cell[0]
        average_orientation_matrix = orientation_matrix[0]

    # Save the unit cell data
    nx_sample["name"] = "FROM_DIALS"
    nx_sample["unit_cell"] = unit_cell
    nx_sample["unit_cell"].attrs["angles_units"] = "deg"
    nx_sample["unit_cell"].attrs["length_units"] = "angstrom"

    # Save the orientation matrix
    nx_sample["orientation_matrix"] = orientation_matrix

    # Set an average unit cell etc for scan static stuff
    nx_sample["average_unit_cell"] = average_unit_cell
    nx_sample["average_unit_cell"].attrs["angles_units"] = "deg"
    nx_sample["average_unit_cell"].attrs["length_units"] = "angstrom"
    nx_sample["average_orientation_matrix"] = average_orientation_matrix

    # Set depends on
    if scan is not None:
        nx_sample["depends_on"] = str(nx_sample["transformations/phi"].name)
    else:
        nx_sample["depends_on"] = "."


def dump_details(entry):
    from time import strftime

    # Program info
    entry["program_name"] = "dials.export_nxmx"
    entry["program_name"].attrs["version"] = 1
    entry["program_name"].attrs["configuration"] = ""

    # Set some processing information (each program should add itself)
    nx_process = get_nx_process(entry, "process")
    nx_process["program"] = "dials"
    nx_process["version"] = 1
    nx_process["date"] = strftime("%Y-%m-%dT%H:%M:%S")

    nx_note = get_nx_note(nx_process, "spot_finding")
    nx_note["author"] = "dials.find_spots"
    nx_note["date"] = strftime("%Y-%m-%dT%H:%M:%S")
    nx_note["type"] = "text/plain"
    nx_note["description"] = "Spot finding parameters"
    nx_note["data"] = "dials.find_spots imported.expt"
    nx_note["sequence_index"] = 0

    nx_note = get_nx_note(nx_process, "indexing")
    nx_note["author"] = "dials.index"
    nx_note["date"] = strftime("%Y-%m-%dT%H:%M:%S")
    nx_note["type"] = "text/plain"
    nx_note["description"] = "Indexing parameters"
    nx_note["data"] = "dials.index imported.expt strong.refl"
    nx_note["sequence_index"] = 1

    nx_note = get_nx_note(nx_process, "refinement")
    nx_note["author"] = "dials.refine"
    nx_note["date"] = strftime("%Y-%m-%dT%H:%M:%S")
    nx_note["type"] = "text/plain"
    nx_note["description"] = "Refinement parameters"
    nx_note["data"] = "dials.refine indexed.expt indexed.refl"
    nx_note["sequence_index"] = 2

    nx_note = get_nx_note(nx_process, "integration")
    nx_note["author"] = "dials.integrate"
    nx_note["date"] = strftime("%Y-%m-%dT%H:%M:%S")
    nx_note["type"] = "text/plain"
    nx_note["description"] = "Integration parameters"
    nx_note["data"] = "dials.integrate refined.expt refined.refl"
    nx_note["sequence_index"] = 3


def load_beam(entry):
    from dxtbx.model import Beam

    EPS = 1e-7

    # Get the nx_beam
    nx_sample = get_nx_sample(entry, "sample")
    nx_beam = get_nx_beam(nx_sample, "beam")
    wavelength = nx_beam["incident_wavelength"][()]
    S0, S1, S2, S3 = tuple(nx_beam["incident_polarization_stokes"])
    n, p = polarization_stokes_to_normal(S0, S1, S2, S3)
    assert n.dot(matrix.col((0, 0, -1))) < EPS

    # Return the beam model
    return Beam((0, 0, -1), wavelength, 0, 0, n, p, 0, 1)


def load_detector(entry):
    from dxtbx.model import Detector

    # Get the detector module object
    nx_instrument = get_nx_instrument(entry, "instrument")
    nx_detector = get_nx_detector(nx_instrument, "detector")
    assert h5str(nx_detector["depends_on"][()]) == "."
    material = nx_detector["sensor_material"][()]
    det_type = nx_detector["type"][()]
    thickness = nx_detector["sensor_thickness"][()]
    trusted_range = (nx_detector["underload"][()], nx_detector["saturation_value"][()])

    # The detector model
    detector = Detector()

    i = 0
    while True:
        try:
            module = get_nx_detector_module(nx_detector, "module%d" % i)
        except Exception:
            break
        # Set the data size
        image_size = module["data_size"]

        # Set the module offset
        offset_length = module["module_offset"][()]
        assert module["module_offset"].attrs["depends_on"] == "."
        assert module["module_offset"].attrs["transformation_type"] == "translation"
        assert tuple(module["module_offset"].attrs["offset"]) == (0, 0, 0)
        offset_vector = matrix.col(module["module_offset"].attrs["vector"])
        origin = offset_vector * offset_length

        # Write the fast pixel direction
        module_offset_path = str(module["module_offset"].name)
        pixel_size_x = module["fast_pixel_direction"][()]
        assert module["fast_pixel_direction"].attrs["depends_on"] == module_offset_path
        assert (
            module["fast_pixel_direction"].attrs["transformation_type"] == "translation"
        )
        assert tuple(module["fast_pixel_direction"].attrs["offset"]) == (0, 0, 0)
        fast_axis = tuple(module["fast_pixel_direction"].attrs["vector"])

        # Write the slow pixel direction
        pixel_size_y = module["slow_pixel_direction"][()]
        assert module["slow_pixel_direction"].attrs["depends_on"] == module_offset_path
        assert (
            module["slow_pixel_direction"].attrs["transformation_type"] == "translation"
        )
        assert tuple(module["slow_pixel_direction"].attrs["offset"]) == (0, 0, 0)
        slow_axis = tuple(module["slow_pixel_direction"].attrs["vector"])

        # Get the pixel size and axis vectors
        pixel_size = (pixel_size_x, pixel_size_y)

        # Create the panel
        panel = detector.add_panel()
        panel.set_frame(fast_axis, slow_axis, origin)
        panel.set_pixel_size(pixel_size)
        panel.set_image_size([int(x) for x in image_size])
        panel.set_type(det_type)
        panel.set_thickness(thickness)
        panel.set_material(material)
        panel.set_trusted_range([float(x) for x in trusted_range])
        i += 1

    # Return the detector and panel
    return detector


def load_goniometer(entry):
    from dxtbx.model import Goniometer

    # Write out the rotation axis and oscillation
    nx_sample = get_nx_sample(entry, "sample")
    try:
        transformations = get_nx_transformations(nx_sample, "transformations")
    except Exception:
        return None
    assert transformations["phi"].attrs["depends_on"] == str(
        transformations["fixed_rotation"].name
    )
    assert transformations["phi"].attrs["transformation_type"] == "rotation"
    assert transformations["phi"].attrs["offset_units"] == "mm"
    assert tuple(transformations["phi"].attrs["offset"]) == (0, 0, 0)
    rotation_axis = tuple(transformations["phi"].attrs["vector"])

    assert transformations["fixed_rotation"].attrs["depends_on"] == str(
        transformations["setting_rotation"].name
    )
    assert transformations["fixed_rotation"].attrs["transformation_type"] == "rotation"
    assert transformations["fixed_rotation"].attrs["offset_units"] == "mm"
    assert tuple(transformations["phi"].attrs["offset"]) == (0, 0, 0)
    axis = matrix.col(transformations["fixed_rotation"].attrs["vector"])
    angle = transformations["fixed_rotation"][()]
    fixed_rotation = axis.axis_and_angle_as_r3_rotation_matrix(angle)

    assert transformations["setting_rotation"].attrs["depends_on"] == "."
    assert (
        transformations["setting_rotation"].attrs["transformation_type"] == "rotation"
    )
    assert transformations["setting_rotation"].attrs["offset_units"] == "mm"
    assert tuple(transformations["phi"].attrs["offset"]) == (0, 0, 0)
    axis = matrix.col(transformations["setting_rotation"].attrs["vector"])
    angle = transformations["setting_rotation"][()]
    setting_rotation = axis.axis_and_angle_as_r3_rotation_matrix(angle)

    # Return the goniometer model
    return Goniometer(rotation_axis, fixed_rotation, setting_rotation)


def load_scan(entry):
    from dxtbx.model import Scan

    # Write out the rotation axis and oscillation
    nx_sample = get_nx_sample(entry, "sample")
    try:
        transformations = get_nx_transformations(nx_sample, "transformations")
    except Exception:
        return None
    phi = transformations["phi"]
    assert transformations["phi"].attrs["transformation_type"] == "rotation"
    assert transformations["phi"].attrs["offset_units"] == "mm"
    assert tuple(transformations["phi"].attrs["offset"]) == (0, 0, 0)
    image_range = (1, len(phi))
    oscillation = (phi[0], phi[1] - phi[0])
    nx_instrument = get_nx_instrument(entry, "instrument")
    nx_detector = get_nx_detector(nx_instrument, "detector")
    exposure_time = nx_detector["frame_time"]
    epochs = nx_detector["timestamp"]
    return Scan(image_range, oscillation, exposure_time, epochs, deg=True)


def load_crystal(entry):
    from cctbx import uctbx
    from dxtbx.model import Crystal
    from scitbx.array_family import flex

    # Get the sample
    nx_sample = get_nx_sample(entry, "sample")

    # Set the space group
    space_group_symbol = nx_sample["unit_cell_group"][()]

    # Get depends on
    if h5str(nx_sample["depends_on"][()]) != ".":
        assert h5str(nx_sample["depends_on"][()]) == h5str(
            nx_sample["transformations/phi"].name
        )

    # Read the average unit cell data
    average_unit_cell = flex.double(np.array(nx_sample["average_unit_cell"]))
    assert nx_sample["average_unit_cell"].attrs["angles_units"] == "deg"
    assert nx_sample["average_unit_cell"].attrs["length_units"] == "angstrom"
    assert len(average_unit_cell.all()) == 1
    assert len(average_unit_cell) == 6
    average_orientation_matrix = flex.double(
        np.array(nx_sample["average_orientation_matrix"])
    )
    assert len(average_orientation_matrix.all()) == 2
    assert average_orientation_matrix.all()[0] == 3
    assert average_orientation_matrix.all()[1] == 3

    # Get the real space vectors
    uc = uctbx.unit_cell(tuple(average_unit_cell))
    U = matrix.sqr(average_orientation_matrix)
    B = matrix.sqr(uc.fractionalization_matrix()).transpose()
    A = U * B
    A = A.inverse()
    real_space_a = A[0:3]
    real_space_b = A[3:6]
    real_space_c = A[6:9]

    # Read the unit cell data
    unit_cell = flex.double(np.array(nx_sample["unit_cell"]))
    assert nx_sample["unit_cell"].attrs["angles_units"] == "deg"
    assert nx_sample["unit_cell"].attrs["length_units"] == "angstrom"

    # Read the orientation matrix
    orientation_matrix = flex.double(np.array(nx_sample["orientation_matrix"]))
    assert len(unit_cell.all()) == 2
    assert len(orientation_matrix.all()) == 3
    assert unit_cell.all()[0] == orientation_matrix.all()[0]
    assert unit_cell.all()[1] == 6
    assert orientation_matrix.all()[1] == 3
    assert orientation_matrix.all()[2] == 3

    # Construct the crystal model
    crystal = Crystal(real_space_a, real_space_b, real_space_c, space_group_symbol)

    # Sort out scan points
    if unit_cell.all()[0] > 1:
        A_list = []
        for i in range(unit_cell.all()[0]):
            uc = uctbx.unit_cell(tuple(unit_cell[i : i + 1, :]))
            U = matrix.sqr(tuple(orientation_matrix[i : i + 1, :, :]))
            B = matrix.sqr(uc.fractionalization_matrix()).transpose()
            A_list.append(U * B)
        crystal.set_A_at_scan_points(A_list)
    else:
        assert unit_cell.all_eq(average_unit_cell)
        assert orientation_matrix.all_eq(average_orientation_matrix)

    # Return the crystal
    return crystal


def dump(entry, experiments, params):
    from dxtbx.imageset import ImageSequence

    print("Dumping NXmx")

    # Rotate the experiments such that beam direction is along (0, 0, -1)
    experiments, rotations = convert_to_nexus_beam_direction(experiments)

    # Add the feature
    if "features" in entry:
        features = entry["features"]
        assert features.dtype == "uint64"
        features.resize((len(features) + 1,))
        features[len(features) - 1] = 6
    else:
        features = entry.create_dataset(
            "features", (1,), maxshape=(None,), dtype=np.uint64
        )
        features[0] = 6

    exp_names = []

    # Get the experiment
    for index, experiment in enumerate(experiments):

        # Create the entry
        assert ("experiment_%d" % index) not in entry
        nxmx = entry.create_group("experiment_%d" % index)
        nxmx.attrs["NX_class"] = "NXsubentry"
        exp_names.append(str(nxmx.name))

        # Get the dials specific stuff
        nx_dials = get_nx_dials(nxmx, "dials")
        nx_dials["index"] = index
        nx_dials["index"].attrs["source"] = experiment.beam_id
        nx_dials["index"].attrs["detector"] = experiment.detector_id
        if experiment.goniometer is not None:
            nx_dials["index"].attrs["goniometer"] = experiment.goniometer_id
        if experiment.scan is not None:
            nx_dials["index"].attrs["scan"] = experiment.scan_id
        nx_dials["index"].attrs["sample"] = experiment.crystal_id

        # Write out the original orientation (dials specific)
        transformations = get_nx_transformations(nx_dials, "transformations")
        transformations["angle"] = -rotations[index][1]
        transformations["angle"].attrs["transformation_type"] = "rotation"
        transformations["angle"].attrs["vector"] = rotations[index][0]
        transformations["angle"].attrs["offset"] = (0, 0, 0)
        transformations["angle"].attrs["offset_units"] = "mm"
        transformations["angle"].attrs["depends_on"] = "."

        # Create the imageset template
        if experiment.imageset is None:
            nx_dials["template"] = ""
            if experiment.scan is not None:
                nx_dials["template"].attrs["range"] = experiment.scan.get_image_range()
        else:
            from os.path import abspath

            if isinstance(experiment.imageset, ImageSequence):
                template = abspath(experiment.imageset.get_template())
                nx_dials["template"] = template
                nx_dials["template"].attrs["range"] = experiment.scan.get_image_range()
            else:
                template = [
                    abspath(experiment.imageset.get_path(i))
                    for i in range(len(experiment.imageset))
                ]
                nx_dials["template"] = template

        # Create the definition
        definition = nxmx.create_dataset("definition", data="NXmx")
        definition.attrs["version"] = 1
        definition.attrs["URL"] = schema_url

        nxmx["title"] = "FROM_DIALS"

        # Dump the models
        dump_beam(nxmx, experiment.beam)
        dump_detector(
            nxmx,
            experiment.detector,
            experiment.beam,
            experiment.imageset,
            experiment.scan,
        )
        dump_goniometer(nxmx, experiment.goniometer, experiment.scan)
        dump_crystal(nxmx, experiment.crystal, experiment.scan)

        # Add instrument name and source
        nx_instrument = get_nx_instrument(nxmx, "instrument")
        nx_instrument["name"] = params.instrument_name
        nx_instrument["name"].attrs["short_name"] = params.instrument_short_name

        nx_source = get_nx_source(nxmx, "source")
        nx_source["name"] = params.source_name
        nx_source["name"].attrs["short_name"] = params.source_short_name

    # Dump some details
    dump_details(entry)

    # Link the data
    # nxmx['data'] = nxmx['instrument/detector/data']

    return exp_names


def find_nx_mx_entries(nx_file, entry):
    """
    Find NXmx entries
    """
    hits = []

    def visitor(name, obj):
        if "NX_class" in obj.attrs:
            if h5str(obj.attrs["NX_class"]) in ["NXentry", "NXsubentry"]:
                if "definition" in obj:
                    if h5str(obj["definition"][()]) == "NXmx":
                        hits.append(obj)

    nx_file[entry].visititems(visitor)
    return hits


def load(entry, exp_index):
    from dxtbx.model.experiment_list import Experiment, ExperimentList

    print("Loading NXmx")

    # Check file contains the feature
    assert "features" in entry
    assert 6 in entry["features"][()]

    experiment_list = ExperimentList()

    # Find all the experiments
    entries = find_nx_mx_entries(entry, ".")
    if len(entries) > 1:
        entries = sorted(entries, key=lambda x: x["dials/index"][()])

    assert len(entries) == len(exp_index)
    for nxmx, name in zip(entries, exp_index):
        assert np.string_(nxmx.name) == name, (nxmx.name, name)

    index = []
    rotations = []
    for name in exp_index:

        # Get the entry
        nxmx = entry.file[name]

        # Get the definition
        definition = nxmx["definition"]
        assert h5str(definition[()]) == "NXmx"
        assert definition.attrs["version"] == 1

        # Get dials specific stuff
        nx_dials = get_nx_dials(nxmx, "dials")

        # Set index
        b = nx_dials["index"].attrs["source"]
        d = nx_dials["index"].attrs["detector"]
        if "goniometer" in nx_dials["index"].attrs:
            g = nx_dials["index"].attrs["goniometer"]
        else:
            g = None
        if "scan" in nx_dials["index"].attrs:
            s = nx_dials["index"].attrs["scan"]
        else:
            s = None
        c = nx_dials["index"].attrs["sample"]
        index.append((b, d, g, s, c))

        # Get the original orientation (dials specific)
        transformations = get_nx_transformations(nx_dials, "transformations")
        angle = transformations["angle"][()]
        assert transformations["angle"].attrs["transformation_type"] == "rotation"
        axis = transformations["angle"].attrs["vector"]
        assert tuple(transformations["angle"].attrs["offset"]) == (0, 0, 0)
        assert transformations["angle"].attrs["offset_units"] == "mm"
        assert transformations["angle"].attrs["depends_on"] == "."
        rotations.append((axis, angle))

        # Get the imageset
        try:
            image_range = nx_dials["template"].attrs["range"]
        except Exception:
            image_range = None

        # Create the experiment
        experiment = Experiment()

        # Read the models
        experiment.beam = load_beam(nxmx)
        experiment.detector = load_detector(nxmx)
        experiment.goniometer = load_goniometer(nxmx)
        experiment.scan = load_scan(nxmx)
        experiment.crystal = load_crystal(nxmx)

        # Set the image range
        if image_range is not None and experiment.scan is not None:
            num = image_range[1] - image_range[0] + 1
            assert num == len(experiment.scan)
            experiment.scan.set_image_range([int(x) for x in image_range])

        # Return the experiment list
        experiment_list.append(experiment)

    # Convert from nexus beam direction
    experiment_list = convert_from_nexus_beam_direction(experiment_list, rotations)

    beam = collections.defaultdict(list)
    detector = collections.defaultdict(list)
    goniometer = collections.defaultdict(list)
    scan = collections.defaultdict(list)
    crystal = collections.defaultdict(list)
    for i, ind in enumerate(index):
        beam[ind[0]].append(i)
        detector[ind[1]].append(i)
        goniometer[ind[2]].append(i)
        scan[ind[3]].append(i)
        crystal[ind[4]].append(i)

    # Set all the shared beams
    for value in beam.values():
        b1 = experiment_list[value[0]].beam
        assert all(experiment_list[v].beam == b1 for v in value[1:])
        for v in value[1:]:
            experiment_list[v].beam = b1
    # Set all the shared detectors
    for value in detector.values():
        d1 = experiment_list[value[0]].detector
        assert all(experiment_list[v].detector == d1 for v in value[1:])
        for v in value[1:]:
            experiment_list[v].detector = d1
    # Set all the shared goniometer
    for value in goniometer.values():
        g1 = experiment_list[value[0]].goniometer
        assert all(experiment_list[v].goniometer == g1 for v in value[1:])
        for v in value[1:]:
            experiment_list[v].goniometer = g1
    # Set all the shared scans
    for value in scan.values():
        s1 = experiment_list[value[0]].scan
        assert all(experiment_list[v].scan == s1 for v in value[1:])
        for v in value[1:]:
            experiment_list[v].scan = s1
    # Set all the shared crystals
    for value in crystal.values():
        c1 = experiment_list[value[0]].crystal
        assert all(experiment_list[v].crystal == c1 for v in value[1:])
        for v in value[1:]:
            experiment_list[v].crystal = c1

    return experiment_list



""""
Classes that each define a smoothly varying component of a scaling model.

These classes use a gaussian smoother (1D, 2D or 3D) to calculate the
inverse scale factors and derivatives with respect to the component
parameters.
"""

from __future__ import annotations

from math import ceil, floor

from scitbx import sparse

from dials.algorithms.scaling.model.components.scale_components import (
    ScaleComponentBase,
)
from dials.array_family import flex
from dials_refinement_helpers_ext import GaussianSmoother2D as GS2D
from dials_refinement_helpers_ext import GaussianSmoother3D as GS3D
from dials_scaling_ext import GaussianSmootherFirstFixed as GS1D
from dials_scaling_ext import row_multiply

# The following gaussian smoother classes make the implementation
# consistent with that used in dials.refinement.


class GaussianSmoother1D(GS1D):
    """A 1D Gaussian smoother."""

    def value_weight(self, x, value):
        """Return the value, weight and sumweight at a single point."""
        result = super().value_weight(x, value)
        return (result.get_value(), result.get_weight(), result.get_sumweight())

    def value_weight_first_fixed(self, x, value):
        """Return the value, weight and sumweight at a single point."""
        result = super().value_weight_first_fixed(x, value)
        return (result.get_value(), result.get_weight(), result.get_sumweight())

    def multi_value_weight(self, x, value):
        """Return the value, weight and sumweight at multiple points."""
        result = super().multi_value_weight(x, value)
        return (result.get_value(), result.get_weight(), result.get_sumweight())

    def multi_value_weight_first_fixed(self, x, value):
        """Return the value, weight and sumweight at multiple points."""
        result = super().multi_value_weight_first_fixed(x, value)
        return (result.get_value(), result.get_weight(), result.get_sumweight())

    def positions(self):
        """Return the smoother positions."""
        return list(super().positions())


class GaussianSmoother2D(GS2D):
    """A 2D Gaussian smoother."""

    def value_weight(self, x, y, value):
        """Return the value, weight and sumweight at a single point."""
        result = super().value_weight(x, y, value)
        return (result.get_value(), result.get_weight(), result.get_sumweight())

    def multi_value_weight(self, x, y, value):
        """Return the value, weight and sumweight at multiple points."""
        result = super().multi_value_weight(x, y, value)
        return (result.get_value(), result.get_weight(), result.get_sumweight())

    def x_positions(self):
        """Return the smoother x-positions."""
        return list(super().x_positions())

    def y_positions(self):
        """Return the smoother y-positions."""
        return list(super().y_positions())


class GaussianSmoother3D(GS3D):
    """A 3D Gaussian smoother."""

    def value_weight(self, x, y, z, value):
        """Return the value, weight and sumweight at a single point."""
        result = super().value_weight(x, y, z, value)
        return (result.get_value(), result.get_weight(), result.get_sumweight())

    def multi_value_weight(self, x, y, z, value):
        """Return the value, weight and sumweight at multiple points."""
        result = super().multi_value_weight(x, y, z, value)
        return (result.get_value(), result.get_weight(), result.get_sumweight())

    def x_positions(self):
        """Return the smoother x-positions."""
        return list(super().x_positions())

    def y_positions(self):
        """Return the smoother y-positions."""
        return list(super().y_positions())

    def z_positions(self):
        """Return the smoother z-positions."""
        return list(super().z_positions())


class SmoothMixin:
    """Mixin class for smooth scale factor components.

    This uses a Gaussian smoother to calculate scales and derivatives
    based on the parameters and a have a set of normalised_values
    associated with the data."""

    def __init__(self):
        self._Vr = 1.0
        self._smoother = None

    @property
    def value(self):
        """Extra access to the parameters for the gaussian smoother."""
        return self._parameters

    @property
    def smoother(self):
        """The Gaussian smoother."""
        return self._smoother

    @staticmethod
    def nparam_to_val(n_params):
        """Convert the number of parameters to the required input value
        for the smoother."""
        assert (
            n_params >= 2
        ), """cannot initialise a smooth scale factor
      for <2 parameters."""
        if n_params == 2 or n_params == 3:
            return n_params - 1
        return n_params - 2


class SmoothScaleComponent1D(ScaleComponentBase, SmoothMixin):
    """A smoothly varying scale component in one dimension.

    This class has the option to fix the first parameter during minimisation."""

    null_parameter_value = 1.0

    def __init__(self, initial_values, parameter_esds=None):
        super().__init__(initial_values, parameter_esds)
        self._normalised_values = []
        self._fixed_initial = False

    def fix_initial_parameter(self):
        """Set a flag to indicate that we're fixing the first parameter."""
        self._fixed_initial = True

    @property
    def free_parameters(self):
        if self._fixed_initial:
            return self._parameters[1:]
        return self._parameters

    @free_parameters.setter
    def free_parameters(self, parameters):
        if not self._fixed_initial:
            self._parameters = parameters
        else:
            sel = flex.bool(self._parameters.size(), True)
            sel[0] = False
            self._parameters.set_selected(sel, parameters)

    @property
    def free_parameter_esds(self):
        """Return the estimated standard deviations of the parameters."""
        if self._fixed_initial:
            return self._parameter_esds[1:]
        return self._parameter_esds

    @free_parameter_esds.setter
    def free_parameter_esds(self, esds):
        assert len(esds) == len(self.free_parameters)
        if not self._fixed_initial:
            self._parameter_esds = esds
        else:
            sel = flex.bool(self._parameters.size(), True)
            sel[0] = False
            if self._parameter_esds:
                self._parameter_esds.set_selected(sel, esds)
                self._parameter_esds[0] = 0.0
            else:
                self._parameter_esds = flex.double(self.parameters.size(), 0.0)
                self._parameter_esds.set_selected(sel, esds)

    def set_new_parameters(self, new_parameters):
        """Set new parameters of a different length i.e. after batch handling"""
        self._parameters = new_parameters
        self._parameter_esds = None
        self._n_params = len(self._parameters)

    @property
    def normalised_values(self):
        """This is a list of the relevant data needed to calculate the
        inverse scale factors, normalised to give 'normalised coordinate
        values' that fit in the range of the smoother parameters, which
        are defined as a 1D array at normalised coordinates separated by
        a spacing of 1."""
        return self._normalised_values

    @ScaleComponentBase.data.setter
    def data(self, data):
        assert set(data.keys()) == {"x"}, set(data.keys())
        self._data = data

    def update_reflection_data(self, selection=None, block_selections=None):
        """Set the normalised coordinate values and configure the smoother."""
        self._normalised_values = []
        self._n_refl = []
        normalised_values = self.data["x"]
        if selection:
            normalised_values = normalised_values.select(selection)
        # Make sure zeroed correctly.
        normalised_values = normalised_values - flex.min(normalised_values)
        phi_range_deg = [
            floor(round(flex.min(normalised_values), 10)),
            max(ceil(round(flex.max(normalised_values), 10)), 1),
        ]
        self._smoother = GaussianSmoother1D(
            phi_range_deg, self.nparam_to_val(self._n_params)
        )
        if block_selections:
            block_selection_list = block_selections
            for i, sel in enumerate(block_selection_list):
                self._normalised_values.append(normalised_values.select(sel))
                self._n_refl.append(self._normalised_values[i].size())
        else:
            self._normalised_values.append(normalised_values)
            self._n_refl.append(normalised_values.size())

    def calculate_scales_and_derivatives(self, block_id=0):
        if self._n_refl[block_id] > 1:
            if self._fixed_initial:
                (
                    value,
                    weight,
                    sumweight,
                ) = self._smoother.multi_value_weight_first_fixed(
                    self._normalised_values[block_id], self.value
                )
            else:
                value, weight, sumweight = self._smoother.multi_value_weight(
                    self._normalised_values[block_id], self.value
                )
            inv_sw = 1.0 / sumweight
            dv_dp = row_multiply(weight, inv_sw)
        elif self._n_refl[block_id] == 1:
            if self._fixed_initial:
                value, weight, sumweight = self._smoother.value_weight_first_fixed(
                    self._normalised_values[block_id][0], self.value
                )
            else:
                value, weight, sumweight = self._smoother.value_weight(
                    self._normalised_values[block_id][0], self.value
                )
            dv_dp = sparse.matrix(1, weight.size)
            b = flex.double(weight.as_dense_vector() / sumweight)
            b.reshape(flex.grid(1, b.size()))
            dv_dp.assign_block(b, 0, 0)
            value = flex.double(1, value)
        else:
            return flex.double([]), sparse.matrix(0, 0)
        return value, dv_dp

    def calculate_scales(self, block_id=0):
        """ "Only calculate the scales if needed, for performance."""
        if self._n_refl[block_id] > 1:
            value, _, __ = self._smoother.multi_value_weight(
                self._normalised_values[block_id], self.value
            )
        elif self._n_refl[block_id] == 1:
            value, _, __ = self._smoother.value_weight(
                self._normalised_values[block_id][0], self.value
            )
            value = flex.double(1, value)
        else:
            value = flex.double([])
        return value


class SmoothBScaleComponent1D(SmoothScaleComponent1D):
    """Subclass of SmoothScaleComponent1D to implement a smoothly
    varying B-factor correction."""

    null_parameter_value = 0.0

    def __init__(self, initial_values, parameter_esds=None):
        super().__init__(initial_values, parameter_esds)
        self._d_values = []

    @property
    def d_values(self):
        """The current set of d-values associated with this component."""
        return self._d_values

    @ScaleComponentBase.data.setter
    def data(self, data):
        assert set(data.keys()) == {"x", "d"}, set(data.keys())
        self._data = data

    def update_reflection_data(self, selection=None, block_selections=None):
        super().update_reflection_data(selection, block_selections)
        self._d_values = []
        data = self.data["d"]
        if selection:
            data = data.select(selection)
        if block_selections:
            for sel in block_selections:
                self._d_values.append(data.select(sel))
        else:
            self._d_values.append(data)

    def calculate_scales_and_derivatives(self, block_id=0):
        scales, derivatives = super().calculate_scales_and_derivatives(block_id)
        if self._n_refl[block_id] == 0:
            return flex.double([]), sparse.matrix(0, 0)
        prefac = 1.0 / (2.0 * (self._d_values[block_id] * self._d_values[block_id]))
        s = flex.exp(scales * prefac)
        d = row_multiply(derivatives, s * prefac)
        return s, d

    def calculate_scales(self, block_id=0):
        s = super().calculate_scales(block_id)
        return flex.exp(s / (2.0 * flex.pow2(self._d_values[block_id])))

    def calculate_restraints(self):
        residual = self.parameter_restraints * (self._parameters * self._parameters)
        gradient = 2.0 * self.parameter_restraints * self._parameters
        return residual, gradient

    def calculate_jacobian_restraints(self):
        jacobian = sparse.matrix(self.n_params, self.n_params)
        for i in range(self.n_params):
            jacobian[i, i] = +1.0
        return self._parameters, jacobian, self.parameter_restraints


class SmoothScaleComponent2D(ScaleComponentBase, SmoothMixin):
    """Implementation of a 2D array-based smoothly varying scale factor.

    A 2d array of parameters is defined, and the scale factor at fractional
    coordinates is calculated as smoothly varying based on the distance to
    the nearby parameters as calculated in the GaussianSmoother2D. The
    initial values are passed as a 1D array, and shape is a 2-tuple
    indicating the number of parameters in each dimension."""

    null_parameter_value = 1.0

    def __init__(self, initial_values, shape, parameter_esds=None):
        assert len(initial_values) == (
            shape[0] * shape[1]
        ), """The shape
    information to initialise a 2D smoother is inconsistent with the length
    of the initial parameter list."""
        super().__init__(initial_values, parameter_esds)
        self._n_x_params = shape[0]
        self._n_y_params = shape[1]
        self._normalised_x_values = None
        self._normalised_y_values = None

    @ScaleComponentBase.data.setter
    def data(self, data):
        assert set(data.keys()) == {"x", "y"}, set(data.keys())
        self._data = data

    def set_new_parameters(self, new_parameters, shape):
        """Set new parameters of a different length i.e. after batch handling"""
        assert len(new_parameters) == shape[0] * shape[1]
        self._parameters = new_parameters
        self._parameter_esds = None
        self._n_params = len(self._parameters)
        self._n_x_params = shape[0]
        self._n_y_params = shape[1]

    @property
    def n_x_params(self):
        """The number of parameters that parameterise the x-component."""
        return self._n_x_params

    @property
    def n_y_params(self):
        """The number of parameters that parameterise the y-component."""
        return self._n_y_params

    @property
    def normalised_x_values(self):
        """The normalised coordinate values in the first dimension."""
        return self._normalised_x_values

    @property
    def normalised_y_values(self):
        """The normalised coordinate values in the second dimension."""
        return self._normalised_y_values

    def update_reflection_data(self, selection=None, block_selections=None):
        """control access to setting all of reflection data at once"""

        self._normalised_x_values = []
        self._normalised_y_values = []
        self._n_refl = []
        normalised_x_values = self.data["x"]
        normalised_y_values = self.data["y"]
        if selection:
            normalised_x_values = normalised_x_values.select(selection)
            normalised_y_values = normalised_y_values.select(selection)
        normalised_x_values = normalised_x_values - flex.min(normalised_x_values)
        normalised_y_values = normalised_y_values - flex.min(normalised_y_values)
        x_range = [
            floor(round(flex.min(normalised_x_values), 10)),
            max(ceil(round(flex.max(normalised_x_values), 10)), 1),
        ]
        y_range = [
            floor(round(flex.min(normalised_y_values), 10)),
            max(ceil(round(flex.max(normalised_y_values), 10)), 1),
        ]
        self._smoother = GaussianSmoother2D(
            x_range,
            self.nparam_to_val(self._n_x_params),
            y_range,
            self.nparam_to_val(self._n_y_params),
        )
        if block_selections:
            for i, sel in enumerate(block_selections):
                self._normalised_x_values.append(normalised_x_values.select(sel))
                self._normalised_y_values.append(normalised_y_values.select(sel))
                self._n_refl.append(self._normalised_x_values[i].size())
        else:
            self._normalised_x_values.append(normalised_x_values)
            self._normalised_y_values.append(normalised_y_values)
            self._n_refl.append(normalised_x_values.size())

    def calculate_scales_and_derivatives(self, block_id=0):
        if self._n_refl[block_id] > 1:
            value, weight, sumweight = self._smoother.multi_value_weight(
                self._normalised_x_values[block_id],
                self._normalised_y_values[block_id],
                self.value,
            )
            inv_sw = 1.0 / sumweight
            dv_dp = row_multiply(weight, inv_sw)
        elif self._n_refl[block_id] == 1:
            value, weight, sumweight = self._smoother.value_weight(
                self._normalised_x_values[block_id][0],
                self._normalised_y_values[block_id][0],
                self.value,
            )
            dv_dp = sparse.matrix(1, weight.size)
            b = flex.double(weight.as_dense_vector() / sumweight)
            b.reshape(flex.grid(1, b.size()))
            dv_dp.assign_block(b, 0, 0)
            value = flex.double(1, value)
        else:
            return flex.double([]), sparse.matrix(0, 0)
        return value, dv_dp

    def calculate_scales(self, block_id=0):
        """Only calculate the scales if needed, for performance."""
        if self._n_refl[block_id] > 1:
            value, _, __ = self._smoother.multi_value_weight(
                self._normalised_x_values[block_id],
                self._normalised_y_values[block_id],
                self.value,
            )
        elif self._n_refl[block_id] == 1:
            value, _, __ = self._smoother.value_weight(
                self._normalised_x_values[block_id][0],
                self._normalised_y_values[block_id][0],
                self.value,
            )
            value = flex.double(1, value)
        else:
            value = flex.double([])
        return value


class SmoothScaleComponent3D(ScaleComponentBase, SmoothMixin):
    """Implementation of a 3D array-based smoothly varying scale factor.

    A 3d array of parameters is defined, and the scale factor at fractional
    coordinates is calculated as smoothly varying based on the distance to
    the nearby parameters as calculated in the GaussianSmoother3D. The
    initial values are passed as a 1D array, and shape is a 3-tuple
    indicating the number of parameters in each dimension."""

    null_parameter_value = 1.0

    def __init__(self, initial_values, shape, parameter_esds=None):
        assert len(initial_values) == (
            shape[0] * shape[1] * shape[2]
        ), """The
    shape information to initialise a 3D smoother is inconsistent with the
    length of the initial parameter list."""
        super().__init__(initial_values, parameter_esds)
        self._n_x_params = shape[0]
        self._n_y_params = shape[1]
        self._n_z_params = shape[2]
        self._normalised_x_values = None
        self._normalised_y_values = None
        self._normalised_z_values = None

    def set_new_parameters(self, new_parameters, shape):
        """Set new parameters of a different length i.e. after batch handling"""
        assert len(new_parameters) == shape[0] * shape[1] * shape[2]
        self._parameters = new_parameters
        self._parameter_esds = None
        self._n_params = len(self._parameters)
        self._n_x_params = shape[0]
        self._n_y_params = shape[1]
        self._n_z_params = shape[2]

    @ScaleComponentBase.data.setter
    def data(self, data):
        assert set(data.keys()) == {"x", "y", "z"}, set(data.keys())
        self._data = data

    @property
    def n_x_params(self):
        """The number of parameters that parameterise the x-component."""
        return self._n_x_params

    @property
    def n_y_params(self):
        """The number of parameters that parameterise the y-component."""
        return self._n_y_params

    @property
    def n_z_params(self):
        """The number of parameters that parameterise the z-component."""
        return self._n_z_params

    @property
    def normalised_x_values(self):
        """The normalised coordinate values in the first dimension."""
        return self._normalised_x_values

    @property
    def normalised_y_values(self):
        """The normalised coordinate values in the second dimension."""
        return self._normalised_y_values

    @property
    def normalised_z_values(self):
        """The normalised coordinate values in the third dimension."""
        return self._normalised_z_values

    def update_reflection_data(self, selection=None, block_selections=None):
        """control access to setting all of reflection data at once"""
        self._normalised_x_values = []
        self._normalised_y_values = []
        self._normalised_z_values = []
        self._n_refl = []
        normalised_x_values = self.data["x"]
        normalised_y_values = self.data["y"]
        normalised_z_values = self.data["z"]
        if selection:
            normalised_x_values = normalised_x_values.select(selection)
            normalised_y_values = normalised_y_values.select(selection)
            normalised_z_values = normalised_z_values.select(selection)
        """Set the normalised coordinate values and configure the smoother."""
        normalised_x_values = normalised_x_values - flex.min(normalised_x_values)
        normalised_y_values = normalised_y_values - flex.min(normalised_y_values)
        normalised_z_values = normalised_z_values - flex.min(normalised_z_values)
        x_range = [
            floor(round(flex.min(normalised_x_values), 10)),
            max(ceil(round(flex.max(normalised_x_values), 10)), 1),
        ]
        y_range = [
            floor(round(flex.min(normalised_y_values), 10)),
            max(ceil(round(flex.max(normalised_y_values), 10)), 1),
        ]
        z_range = [
            floor(round(flex.min(normalised_z_values), 10)),
            max(ceil(round(flex.max(normalised_z_values), 10)), 1),
        ]
        self._smoother = GaussianSmoother3D(
            x_range,
            self.nparam_to_val(self._n_x_params),
            y_range,
            self.nparam_to_val(self._n_y_params),
            z_range,
            self.nparam_to_val(self._n_z_params),
        )
        if block_selections:
            for i, sel in enumerate(block_selections):
                self._normalised_x_values.append(normalised_x_values.select(sel))
                self._normalised_y_values.append(normalised_y_values.select(sel))
                self._normalised_z_values.append(normalised_z_values.select(sel))
                self._n_refl.append(self._normalised_x_values[i].size())
        else:
            self._normalised_x_values.append(normalised_x_values)
            self._normalised_y_values.append(normalised_y_values)
            self._normalised_z_values.append(normalised_z_values)
            self._n_refl.append(normalised_x_values.size())

    def calculate_scales_and_derivatives(self, block_id=0):
        if self._n_refl[block_id] > 1:
            value, weight, sumweight = self._smoother.multi_value_weight(
                self._normalised_x_values[block_id],
                self._normalised_y_values[block_id],
                self._normalised_z_values[block_id],
                self.value,
            )
            inv_sw = 1.0 / sumweight
            dv_dp = row_multiply(weight, inv_sw)
        elif self._n_refl[block_id] == 1:
            value, weight, sumweight = self._smoother.value_weight(
                self._normalised_x_values[block_id][0],
                self._normalised_y_values[block_id][0],
                self._normalised_z_values[block_id][0],
                self.value,
            )
            dv_dp = sparse.matrix(1, weight.size)
            b = flex.double(weight.as_dense_vector() / sumweight)
            b.reshape(flex.grid(1, b.size()))
            dv_dp.assign_block(b, 0, 0)
            value = flex.double(1, value)
        else:
            return flex.double([]), sparse.matrix(0, 0)
        return value, dv_dp

    def calculate_scales(self, block_id=0):
        """ "Only calculate the scales if needed, for performance."""
        if self._n_refl[block_id] > 1:
            value, _, __ = self._smoother.multi_value_weight(
                self._normalised_x_values[block_id],
                self._normalised_y_values[block_id],
                self._normalised_z_values[block_id],
                self.value,
            )
        elif self._n_refl[block_id] == 1:
            value, _, __ = self._smoother.value_weight(
                self._normalised_x_values[block_id][0],
                self._normalised_y_values[block_id][0],
                self._normalised_z_values[block_id][0],
                self.value,
            )
            value = flex.double(1, value)
        else:
            value = flex.double([])
        return value



from __future__ import annotations

import functools

from dxtbx.model import ExperimentList

import dials.extensions
from dials.algorithms.integration.integrator import (
    Parameters,
    _finalize_stills,
    _initialize_stills,
)
from dials.algorithms.integration.ssx.ssx_integrate import (
    OutputCollector,
    SimpleIntegrator,
)
from dials.algorithms.profile_model.factory import ProfileModelFactory
from dials.algorithms.shoebox import MaskCode
from dials.array_family import flex
from dials.command_line.integrate import process_reference


class StillsOutputCollector(OutputCollector):
    def collect_after_integration(self, experiment, reflection_table):
        super().collect_after_integration(experiment, reflection_table)
        self.data["profile_model"] = experiment.profile.to_dict()
        self.data["profile_model_mosaicity"] = {
            "b": experiment.profile.sigma_b(deg=True)
        }


class StillsIntegrator(SimpleIntegrator):
    def __init__(self, params, collect_data=False):
        super().__init__(params)
        if collect_data:
            self.collector = StillsOutputCollector()

    def run(self, experiment, table):

        # first set ids to zero so can integrate (this is how integration
        # finds the image in the imageset)
        ids_map = dict(table.experiment_identifiers())
        table["id"] = flex.int(table.size(), 0)
        del table.experiment_identifiers()[list(ids_map.keys())[0]]
        table.experiment_identifiers()[0] = list(ids_map.values())[0]

        self.collector.initial_collect(experiment, table)

        # preprocess
        table = self.preprocess(table)
        self.collector.collect_after_preprocess(experiment, table)
        # refine - null

        predicted, elist = self.predict(experiment, table, self.params)
        self.collector.collect_after_prediction(predicted, table)
        integrated, elist = self.integrate(elist, predicted, self.params)
        # NB what about things like kapton correction?
        self.collector.collect_after_integration(elist[0], integrated)

        return elist[0], integrated, self.collector

    @staticmethod
    def preprocess(table):
        reference, _ = process_reference(table)
        return reference

    @staticmethod
    def refine():
        pass

    @staticmethod
    def predict(experiment, table, params):
        elist = ExperimentList([experiment])
        predicted = flex.reflection_table.from_predictions_multi(
            elist,
            dmin=params.prediction.d_min,
            dmax=params.prediction.d_max,
            margin=params.prediction.margin,
            force_static=params.prediction.force_static,
            padding=params.prediction.padding,
        )
        # need to set imageset id?
        elist = ProfileModelFactory.create(params, elist, table)
        predicted.compute_bbox(elist)
        return predicted, elist

    @staticmethod
    def integrate(experiments, table, params):

        _params = Parameters.from_phil(params.integration)
        experiments[0].scan = None
        BackgroundAlgorithm = dials.extensions.Background.load(
            params.integration.background.algorithm
        )
        flex.reflection_table.background_algorithm = functools.partial(
            BackgroundAlgorithm, params
        )
        CentroidAlgorithm = dials.extensions.Centroid.load(
            params.integration.centroid.algorithm
        )
        flex.reflection_table.centroid_algorithm = functools.partial(
            CentroidAlgorithm, params
        )
        _initialize_stills(experiments, _params, table)

        table["shoebox"] = flex.shoebox(table["panel"], table["bbox"], allocate=True)
        table.extract_shoeboxes(experiments[0].imageset)

        # From integratorexecutor
        table.is_overloaded(experiments)
        table.compute_mask(experiments)
        # Check for invalid pixels in foreground/background
        table.contains_invalid_pixels()
        sbox = table["shoebox"]
        nvalfg = sbox.count_mask_values(MaskCode.Valid | MaskCode.Foreground)
        nforeg = sbox.count_mask_values(MaskCode.Foreground)
        fraction_valid = nvalfg.as_double() / nforeg.as_double()
        selection = (
            fraction_valid < params.integration.profile.valid_foreground_threshold
        )
        table.set_flags(selection, table.flags.dont_integrate)

        # Process the data
        table.compute_background(experiments)
        table.compute_centroid(experiments)
        table.compute_summed_intensity()
        table["num_pixels.valid"] = sbox.count_mask_values(MaskCode.Valid)
        table["num_pixels.background"] = sbox.count_mask_values(
            MaskCode.Valid | MaskCode.Background
        )
        table["num_pixels.background_used"] = sbox.count_mask_values(
            MaskCode.Valid | MaskCode.Background | MaskCode.BackgroundUsed
        )
        table["num_pixels.foreground"] = nvalfg
        table, experiments = _finalize_stills(table, experiments, _params)
        return table, experiments



# DIALS_ENABLE_COMMAND_LINE_COMPLETION


from __future__ import annotations

import concurrent.futures
import copy
import logging
import sys

import iotbx.phil
from dxtbx.model.experiment_list import ExperimentList

from dials.algorithms.indexing import DialsIndexError, indexer
from dials.array_family import flex
from dials.util import log, show_mail_handle_errors
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files
from dials.util.slice import slice_reflections
from dials.util.version import dials_version

logger = logging.getLogger("dials.command_line.index")


help_message = """

This program attempts to perform autoindexing on strong spots output by the
program dials.find_spots. The program is called with a "imported.expt" file
(as generated by dials.import) and a "strong.refl" file (as generated by
dials.find_spots). If one or more lattices are identified given the input
list of strong spots, then the crystal orientation and experimental geometry
are refined to minimise the differences between the observed and predicted
spot centroids. The program will output an "indexed.expt" file which
is similar to the input "imported.expt" file, but with the addition of the
crystal model(s), and an "indexed.refl" file which is similar to the input
"strong.refl" file, but with the addition of miller indices and predicted
spot centroids.

dials.index provides both one-dimensional and three-dimensional fast Fourier
transform (FFT) based methods. These can be chosen by setting the parameters
indexing.method=fft1d or indexing.method=fft3d. By default the program searches
for a primitive lattice, and then proceeds with refinement in space group P1.
If the unit_cell and space_group parameters are set, then the program will
only accept solutions which are consistent with these parameters. Space group
constraints will be enforced in refinement as appropriate.

Examples::

  dials.index imported.expt strong.refl

  dials.index imported.expt strong.refl unit_cell=37,79,79,90,90,90 space_group=P43212

  dials.index imported.expt strong.refl indexing.method=fft1d
"""


phil_scope = iotbx.phil.parse(
    """\
include scope dials.algorithms.indexing.indexer.phil_scope

indexing {

    include scope dials.algorithms.indexing.lattice_search.basis_vector_search_phil_scope

    image_range = None
      .help = "Range in images to slice a sequence. The number of arguments"
              "must be a factor of two. Each pair of arguments gives a range"
              "that follows C conventions (e.g. j0 <= j < j1) when slicing the"
              "reflections by observed centroid."
      .type = ints(size=2)
      .multiple = True

    joint_indexing = True
      .type = bool

}

include scope dials.algorithms.refinement.refiner.phil_scope

output {
  experiments = indexed.expt
    .type = path
  reflections = indexed.refl
    .type = path
  log = dials.index.log
    .type = str
}
""",
    process_includes=True,
)

# override default refinement parameters
phil_overrides = phil_scope.fetch(
    source=iotbx.phil.parse(
        """\
refinement {
    reflections {
        reflections_per_degree=100
    }
}
"""
    )
)

working_phil = phil_scope.fetch(sources=[phil_overrides])


def _index_experiments(experiments, reflections, params, known_crystal_models=None):
    idxr = indexer.Indexer.from_parameters(
        reflections,
        experiments,
        known_crystal_models=known_crystal_models,
        params=params,
    )
    idxr.index()
    idx_refl = copy.deepcopy(idxr.refined_reflections)
    idx_refl.extend(idxr.unindexed_reflections)
    return idxr.refined_experiments, idx_refl


def index(experiments, reflections, params):
    """
    Index the input experiments and reflections.

    Args:
        experiments: The experiments to index
        reflections (list): A list of reflection tables containing strong spots
        params: An instance of the indexing phil scope

    Returns:
        (tuple): tuple containing:
            experiments: The indexed experiment list
            reflections (dials.array_family.flex.reflection_table):
                The indexed reflections

    Raises:
        ValueError: `reflections` is an empty list or `experiments` contains a
                    combination of sequence and stills data.
        dials.algorithms.indexing.DialsIndexError: Indexing failed.
    """
    if experiments.crystals()[0] is not None:
        known_crystal_models = experiments.crystals()
    else:
        known_crystal_models = None

    if len(reflections) == 0:
        raise ValueError("No reflection lists found in input")
    elif len(reflections) == 1:
        if "imageset_id" not in reflections[0]:
            reflections[0]["imageset_id"] = reflections[0]["id"]
    elif len(reflections) > 1:
        assert len(reflections) == len(experiments)
        for i in range(len(reflections)):
            reflections[i]["imageset_id"] = flex.int(len(reflections[i]), i)
            if i > 0:
                reflections[0].extend(reflections[i])
    reflections = reflections[0]

    if params.indexing.image_range:
        reflections = slice_reflections(reflections, params.indexing.image_range)

    if len(experiments) == 1 or params.indexing.joint_indexing:
        indexed_experiments, indexed_reflections = _index_experiments(
            experiments,
            reflections,
            copy.deepcopy(params),
            known_crystal_models=known_crystal_models,
        )
    else:
        indexed_experiments = ExperimentList()

        with concurrent.futures.ProcessPoolExecutor(
            max_workers=params.indexing.nproc
        ) as pool:
            futures = []
            for i_expt, expt in enumerate(experiments):
                refl = reflections.select(reflections["imageset_id"] == i_expt)
                refl["imageset_id"] = flex.size_t(len(refl), 0)
                futures.append(
                    pool.submit(
                        _index_experiments,
                        ExperimentList([expt]),
                        refl,
                        copy.deepcopy(params),
                        known_crystal_models=known_crystal_models,
                    )
                )
            tables_list = []
            for future in concurrent.futures.as_completed(futures):
                try:
                    idx_expts, idx_refl = future.result()
                except Exception as e:
                    print(e)
                else:
                    if idx_expts is None:
                        continue
                    # Update the experiment ids by incrementing by the number of indexed
                    # experiments already in the list
                    ##FIXME below, is i_expt correct - or should it be the
                    # index of the 'future'?
                    idx_refl["imageset_id"] = flex.size_t(idx_refl.size(), i_expt)
                    tables_list.append(idx_refl)
                    indexed_experiments.extend(idx_expts)
        indexed_reflections = flex.reflection_table.concat(tables_list)
    return indexed_experiments, indexed_reflections


@show_mail_handle_errors()
def run(args=None, phil=working_phil):
    usage = "dials.index [options] models.expt strong.refl"

    parser = ArgumentParser(
        usage=usage,
        phil=phil,
        read_reflections=True,
        read_experiments=True,
        check_format=False,
        epilog=help_message,
    )

    params, options = parser.parse_args(args=args, show_diff_phil=False)

    # Configure the logging
    log.config(verbosity=options.verbose, logfile=params.output.log)
    logger.info(dials_version())

    # Log the diff phil
    diff_phil = parser.diff_phil.as_str()
    if diff_phil != "":
        logger.info("The following parameters have been modified:\n")
        logger.info(diff_phil)

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    if len(experiments) == 0:
        parser.print_help()
        return

    try:
        indexed_experiments, indexed_reflections = index(
            experiments, reflections, params
        )
    except (DialsIndexError, ValueError) as e:
        sys.exit(str(e))

    # Save experiments
    logger.info("Saving refined experiments to %s", params.output.experiments)
    assert indexed_experiments.is_consistent()
    indexed_experiments.as_file(params.output.experiments)

    # Save reflections
    logger.info("Saving refined reflections to %s", params.output.reflections)
    indexed_reflections.as_file(filename=params.output.reflections)


if __name__ == "__main__":
    run()
